import numpy as np
from numpy.linalg import norm
from copy import deepcopy

try:
    import trimesh
    MESH_MODE = True
except ImportError:
    MESH_MODE = False

from scipy import optimize
from scipy.signal import argrelmin
from scipy.spatial import KDTree
from scipy.spatial.distance import cdist

from .mesh import WellMesh
from .survey import Survey, _interpolate_survey, slice_survey
from .utils import NEV_to_HLA


class Clearance:
    """
    Initialize a `welleng.clearance.Clearance` object.

    Parameters
    ----------
    reference : `welleng.survey.Survey` object
        The current well from which other wells are referenced.
    offset : `welleng.survey.Survey` object
        The other well.
    k : float
        The dimensionless scaling factor that determines the probability
        of well crossing.
    sigma_pa : float
        Quantifies the 1-SD uncertainty in the projection ahead of the
        current survey station. Its value is partially correlated with
        the projection distance, determined as the current survey depth to
        the bit plus the next survey interval. The magnitude of the actual
        uncertainty also depends on the planned curvature and on the actual
        BHA performance at the wellbore attitude in the formation being
        drilled. The project-ahead uncertainty is only an approximation,
        and although it is predominantly oriented normal to the reference
        well, it is mathematically convenient to define sigma_pa as being
        the radius of a sphere.
    Sm : float
        The surface margin term increases the effective radius of the
        offset well. It accommodates small, unidentified errors and helps
        overcome one of the geometric limitations of the separation rule,
        described in the Separation-Rule Limitations section. It also
        defines the minimum acceptable slot separation during facility
        design and ensures that the separation rule will prohibit the
        activity before nominal contact between the reference and offset
        wells, even if the position uncertainty is zero.
    Rr : float
        The openhole radius of the reference borehole (in meters).
    Ro : float
        The openhole radius of the offset borehole (in meters).
    kop_depth: float
        The kick-off point (measured) depth along the well bore - the
        default value assures that the first survey station is utilized.

    References
    ----------
    Sawaryn, S. J., Wilson, H.. , Bang, J.. , Nyrnes, E.. , Sentance,
    A.. , Poedjono, B.. , Lowdon, R.. , Mitchell, I.. , Codling, J.. ,
    Clark, P. J., and W. T. Allen. "Well-Collision-Avoidance Separation
    Rule." SPE Drill & Compl 34 (2019): 01–15.
    doi: https://doi.org/10.2118/187073-PA
    """
    def __init__(
        self,
        reference: Survey,
        offset: Survey,
        k: float = 3.5,
        sigma_pa: float = 0.5,
        Sm: float = 0.3,
        Rr: float = 0.4572,
        Ro: float = 0.3048,
        kop_depth: float = -np.inf,
    ):
        self.reference = reference
        self.offset = offset
        self.k = k
        self.sigma_pa = sigma_pa
        self.Sm = Sm
        # self.N_verts = N_verts

        self._get_kop_index(kop_depth)
        self._get_ref()
        self._get_radii(Rr, Ro)

        self.ref_nevs = self._get_nevs(self.ref)
        self.offset_nevs = self._get_nevs(self.offset)

    def _get_kop_index(self, kop_depth):
        self.kop_depth = kop_depth
        self.kop_index = np.searchsorted(
            self.reference.md, self.kop_depth, side="left"
        )

    def _get_nevs(self, survey):
        # TODO:
        # - [ ] Take this from the `Survey` where it is already calculated.
        return np.array([
            survey.n,
            survey.e,
            survey.tvd
        ]).T

    def _get_radii(self, Rr, Ro):
        """
        If well bore radius data is provided, use this, otherwise use the
        defaults values.
        """
        if self.ref.radius is not None:
            self.Rr = self.ref.radius
        else:
            self.Rr = np.full(len(self.ref.md), Rr)

        if self.offset.radius is not None:
            self.Ro = self.offset.radius
        else:
            self.Ro = np.full(len(self.offset.md), Ro)

    def _get_ref(self):
        if self.kop_index == 0:
            self.ref = self.reference
        else:
            try:
                cov_nev = self.reference.cov_nev[self.kop_index:]
                cov_hla = self.reference.cov_hla[self.kop_index:]
            except IndexError:
                cov_nev, cov_hla = (None, None)
            self.ref = Survey(
                md=self.reference.md[self.kop_index:],
                inc=self.reference.inc_rad[self.kop_index:],
                azi=self.reference.azi_grid_rad[self.kop_index:],
                n=self.reference.n[self.kop_index:],
                e=self.reference.e[self.kop_index:],
                tvd=self.reference.tvd[self.kop_index:],
                vec=self.reference.vec_nev[self.kop_index:],
                radius=self.reference.radius[self.kop_index:],
                header=self.reference.header,
                cov_nev=cov_nev,
                cov_hla=cov_hla,
                error_model=self.reference.error_model,
                start_xyz=[
                    self.reference.x[self.kop_index],
                    self.reference.y[self.kop_index],
                    self.reference.z[self.kop_index],
                    ],
                start_nev=[
                    self.reference.n[self.kop_index],
                    self.reference.e[self.kop_index],
                    self.reference.tvd[self.kop_index],
                    ],
                deg=self.reference.deg,
                unit=self.reference.unit,
                nev=True
            )


class IscwsaClearance(Clearance):
    """
    Parameters:
    -----------
    clearance_args: List
        See 'welleng.clearance.Clearance` for args.
    minimize_sf: bool
        If `True` (default), then the closest points on the reference well
        are determined and added to the `ref` object as interpolated stations.
    clearance_kwargs: dict
         See 'welleng.clearance.Clearance` for kwargs.

    Attributes:
    -----------
    Ro : array of floats
        The radius of the offset well at each station of the off well.
    Rr : array
        The radius of the reference well at each station on the ref well.
    sf : array of floats
        The calculated Separation Factor to the closest point on the offset
        well for each station on the reference well.
    Sm : float
        The surface margin term increases the effective radius of the
        offset well. It accommodates small, unidentified errors and helps
        overcome one of the geometric limitations of the separation rule,
        described in the Separation-Rule Limitations section. It also
        defines the minimum acceptable slot separation during facility
        design and ensures that the separation rule will prohibit the
        activity before nominal contact between the reference and offset
        wells, even if the position uncertainty is zero.
    calc_hole: array of floats
        The calculated combined equivalent radius of the two well bores, i.e.
        the sum or their radii plus margins.
    closest:
        The closest point on the `off` well from each station on the ref well.
    distance_cc:
        The closest center to center distance for each station on the `ref`
        well to the `off` well.
    eou_boundary:
        The sum of the ellipse of uncertainty radii of the `ref` and `off`
        wells.
    eou_separation:
        The distance between the ellipses of uncertainty of the `ref` and `off`
        wells.
    hoz_bearing:
        The horizontal bearing between the closest points in radians.
    hoz_bearing_deg:
        The horizontal bearing between the closest points in degrees.
    idx: int
        The index of the closest point on the `off` well for each station on
        the `ref` well.
    masd:
        The Minimum Allowable Separation Distance from the `ref` well.
    off: Survey
        The offset well `Survey`.
    off_pcr:
        The Pedal Curve Radii for each station on the `off` well.
    off_cov_hla:
        The covariance matrix in the HLA domain for each station of the `off`
        well.
    off_cov_nev:
        The covariance matrix in the NEV domain for each station of the `off`
        well.
    off_nevs:
        The NEV coordinates of the `off` well.
    offset: Survey
        The initial `offset` well `Survey`.
    offset_nevs:
        The initial NEV coordinates of the `offset` well.
    ref: Survey
        The `ref` well `Survey`.
    ref_pcr:
        The Pedal Curve Radii for each station on the `ref` well.
    ref_cov_hla:
        The covariance matrix in the HLA domain for each station of the `ref`
        well.
    ref_cov_nev:
        The covariance matrix in the NEV domain for each station of the `ref`
        well.
    ref_nevs:
        The NEV coordinates of the `ref` well.
    reference: Survey
        The initial `reference` well `Survey`.
    reference_nevs:
        The initial NEV coordinates of the `reference` well.
    sf:
        The Separation Factor between the closest point on the `off` well for
        each station on the `ref` well.
    toolface_bearing:
        The toolface bearing in radians from each station on the `ref` well to
        the closest point on the `off` well.
    trav_cyl_azi_deg:
        The heading in degrees from each station on teh `ref` well to the
        closest point on the `off` well.
    wellbore_separation:
        The distance between the edge of the wellbore for each station on the
        `ref` well to the closest point on the `off` well.
    """
    def __init__(
        self,
        *clearance_args,
        minimize_sf=None,
        **clearance_kwargs
    ):
        # TODO:
        # - [ ] Can probably remove the `offset` Survey since `off` is a copy.
        # - [ ] Can probably remover the `_nev*` attrs and instead reference
        # the onces in the `Survey` instances.

        super().__init__(*clearance_args, **clearance_kwargs)

        minimize_sf = True if minimize_sf is None else minimize_sf

        # get closest survey station in offset well for each survey
        # station in the reference well
        self.idx = np.argmin(
            cdist(
                self.ref_nevs, self.offset_nevs
            ), axis=-1
        )

        # iterate to find closest point on offset well between
        # survey stations
        self._get_closest_points()
        self.off_nevs = self._get_nevs(self.off)

        # get the unit vectors and horizontal bearing between the wells
        self._get_delta_nev_vectors()

        # transform to HLA coordinates
        # self._get_delta_hla_vectors()

        # make the covariance matrices
        self._get_covs()

        # get the PCRs
        self._get_PCRs()

        # calculate sigmaS
        self.sigmaS = np.sqrt(self.ref_pcr ** 2 + self.off_pcr ** 2)

        # calculate combined hole radii
        self._get_calc_hole()

        # calculate Ellipse of Uncertainty Boundary
        self.eou_boundary = (
            self.k * np.sqrt(self.sigmaS ** 2 + self.sigma_pa ** 2)
        )

        # calculate distance between well bores
        self.wellbore_separation = (
            self.distance_cc.T - self.calc_hole - self.Sm
        )

        # calculate the Ellipse of Uncertainty Separation
        self.eou_separation = (
            self.wellbore_separation - self.eou_boundary
        )

        # calculate the Minimum Allowable Separation Distance
        self.masd = (
            self.eou_boundary + self.calc_hole + self.Sm
        )

        self.sf = np.array(
            self.wellbore_separation / self.eou_boundary,
        )

        # check for minima
        if minimize_sf:
            self.get_sf_mins()

        # for debugging
        # self.pc_method()

    def _get_sf_min(self, x, i, delta_md):
        if x == 0.0:
            return self.sf[i]
        if x == -delta_md[0]:
            return self.sf[i-1]
        if x == delta_md[1]:
            return self.sf[i+1]

        if x < 0:
            ii = i - 1
            xx = delta_md[0] + x
            mult = xx / delta_md[0]
        else:
            ii = i
            xx = x
            mult = xx / delta_md[1]

        node = self.ref.interpolate_md(
            self.ref.md[ii] + xx
        )

        cov_nev = (
            self.ref.cov_nev[ii]
            + (
                np.full(shape=(1, 3, 3), fill_value=mult)
                * (self.ref.cov_nev[ii+1] - self.ref.cov_nev[ii])
            )
        ).reshape(-1, 3, 3)

        sh = self.ref.header
        sh.azi_reference = 'grid'

        survey = Survey(
            md=np.insert(
                self.ref.md[ii: ii+2], 1, node.md
            ),
            inc=np.insert(
                self.ref.inc_rad[ii: ii+2], 1, node.inc_rad
            ),
            azi=np.insert(
                self.ref.azi_grid_rad[ii: ii+2], 1, node.azi_rad
            ),
            cov_nev=np.insert(
                self.ref.cov_nev[ii: ii+2], 1, cov_nev, axis=0
            ),
            start_nev=self.ref.pos_nev[ii],
            deg=False
        )

        clearance_args = (survey, self.offset)
        clearance_kwargs = dict(
            k=self.k,
            sigma_pa=self.sigma_pa,
            Sm=0.0,
            Rr=np.insert(
                self.Rr[ii: ii+2], 1, self.Rr[ii+1]
            ),
            Ro=self.Ro,
            kop_depth=self.kop_depth
        )

        sf_interpolated = IscwsaClearance(
            *clearance_args, **clearance_kwargs, minimize_sf=False
        ).sf[1]

        return sf_interpolated

    def get_sf_mins(self):
        """
        Method for assessing whether a minima has occurred between survey
        station SF values on the reference well and if so calculates the
        minimum SF value between stations (between the previous and next
        station relative to the identified station).

        Modifies the `sf` attribute to include the interpolated minimum `sf`
        values.
        """
        minimas = argrelmin(self.sf)

        sf_interpolated = []

        for minima in minimas[0].tolist():
            delta_md = self.ref.delta_md[minima: minima + 2]
            bounds = [[-delta_md[0], delta_md[1]]]
            # x0 = (np.diff(bounds) / 2)
            x0 = [0]
            args = (minima, delta_md)
            # options = {
            #     'eps': np.sum(delta_md) / 10
            # }

            # SLSQP and L-BFGS-B don't work when using neg to pos ranges in
            # this example, but Powell seems to do the job.
            result = optimize.minimize(
                self._get_sf_min,
                x0,
                method='Powell',
                bounds=bounds,
                args=args,
            )

            if any((
                result.x == 0,
                result.x in delta_md
            )):
                continue

            else:
                sf_interpolated.append((
                    self.ref.md[minima] + result.x[0], result.fun
                ))

        if bool(sf_interpolated):
            for md, sf in sf_interpolated:
                # i = np.searchsorted(self.sf[:, 0], md, side='right')
                i = np.searchsorted(self.ref.md, md, side='right')
                self.sf = np.insert(
                    self.sf, i, np.array([md, sf, 1]), axis=0
                )

                node = self.ref.interpolate_md(md)

                sh = self.ref.header
                sh.azi_reference = 'grid'

                survey = Survey(
                    md=np.insert(
                        self.ref.md, i, node.md
                    ),
                    inc=np.insert(
                        self.ref.inc_rad, i, node.inc_rad
                    ),
                    azi=np.insert(
                        self.ref.azi_grid_rad, i, node.azi_rad
                    ),
                    cov_nev=np.insert(
                        self.ref.cov_nev, i, node.cov_nev, axis=0
                    ),
                    start_nev=self.ref.start_nev,
                    start_xyz=self.ref.start_xyz,
                    deg=False,
                    interpolated=np.insert(
                        self.ref.interpolated, i, True
                    ),
                    radius=np.insert(
                        self.ref.radius, i, self.ref.radius[i]
                    ),
                    header=sh
                )

                self.ref = survey

                pass

            clearance_args = (deepcopy(self.ref), self.offset)
            clearance_kwargs = dict(
                k=self.k,
                sigma_pa=self.sigma_pa,
                Sm=self.Sm,
                Rr=self.Rr,
                Ro=self.Ro,
                kop_depth=self.kop_depth
            )

            # Recalculate all the class lists for the interpolated results
            clearance = IscwsaClearance(
                *clearance_args, **clearance_kwargs, minimize_sf=False
            )

            # Set self vars to clearance vars
            for k, v in clearance.__dict__.items():
                setattr(self, k, v)

            pass

    def get_lines(self):
        """
        Extracts the closest points between wells for each survey section.
        """
        points = [
            [
                c[0],
                [
                    c[1].n[1],
                    c[1].e[1],
                    c[1].tvd[1]
                ]
            ]
            for c in self.closest
        ]

        start_points = []
        end_points = []

        [(start_points.append(p[0]), end_points.append(p[1])) for p in points]

        return np.array([
            np.vstack(start_points).tolist(), np.vstack(end_points).tolist()
        ])

    def _get_closest_points(self):
        """
        Determines the point between pairs of subsequent survey stations on
        the offset well that is closest to each survey stations on the
        reference well.
        """
        closest = []
        for j, (i, station) in enumerate(zip(
            self.idx, self.ref_nevs.tolist()
        )):
            if i > 0:
                bnds = [(0, self.offset.md[i] - self.offset.md[i - 1])]
                res_1 = optimize.minimize(
                    self._fun,
                    bnds[0][1],
                    # method='SLSQP',
                    method='Powell',
                    bounds=bnds,
                    args=(self.offset, i-1, station)
                    )
                mult = res_1.x[0] / (bnds[0][1] - bnds[0][0])
                sigma_new_1 = self._interpolate_covs(i, mult)
            else:
                res_1 = False

            if i < len(self.offset_nevs) - 1:
                bnds = [(0, self.offset.md[i + 1] - self.offset.md[i])]
                res_2 = optimize.minimize(
                    self._fun,
                    bnds[0][0],
                    # method='SLSQP',
                    method='Powell',
                    bounds=bnds,
                    args=(self.offset, i, station)
                    )
                mult = res_2.x[0] / (bnds[0][1] - bnds[0][0])
                sigma_new_2 = self._interpolate_covs(i + 1, mult)
            else:
                res_2 = False

            if res_1 and res_2 and res_1.fun < res_2.fun or not res_2:
                closest.append((
                    station,
                    _interpolate_survey(self.offset, res_1.x[0], i - 1),
                    res_1, sigma_new_1
                ))
            else:
                closest.append((
                    station,
                    _interpolate_survey(self.offset, res_2.x[0], i),
                    res_2,
                    sigma_new_2
                ))

        self.closest = closest
        md, inc, azi, n, e, tvd, x, y, z,  = np.array([
            [
                r[1].md[1],
                r[1].inc_rad[1],
                r[1].azi_grid_rad[1],
                r[1].n[1],
                r[1].e[1],
                r[1].tvd[1],
                r[1].x[1],
                r[1].y[1],
                r[1].z[1],
            ]
            for r in self.closest
        ]).T

        cov_hla = np.array([
            [
                r[3][0],
            ]
            for r in self.closest
        ])

        cov_nev = np.array([
            [
                r[3][1],
            ]
            for r in self.closest
        ])

        self.off = Survey(
            md=md,
            inc=inc,
            azi=azi,
            n=n,
            e=e,
            tvd=tvd,
            header=self.offset.header,
            error_model=None,
            cov_hla=cov_hla,
            cov_nev=cov_nev,
            start_xyz=[x[0], y[0], z[0]],
            start_nev=[n[0], e[0], tvd[0]],
            deg=False,
            unit=self.offset.unit
        )

    def _interpolate_covs(self, i, mult):
        """
        Returns the interpolated covariance matrices for the interpolated
        survey points representing the closest points on the offset well
        relative to each reference well survey station.
        """
        cov_hla_new = (
            self.offset.cov_hla[i - 1]
            + mult * (self.offset.cov_hla[i] - self.offset.cov_hla[i-1])
        )

        cov_nev_new = (
            self.offset.cov_nev[i - 1]
            + mult * (self.offset.cov_nev[i] - self.offset.cov_nev[i-1])
        )

        return (cov_hla_new, cov_nev_new)

    def _fun(self, x, survey, index, station):
        """
        Optimization function used to find the closest point between pairs of
        offset well survey stations.
        """
        s = _interpolate_survey(survey, x[0], index)
        new_pos = np.array([s.n, s.e, s.tvd]).T[1]
        dist = norm(new_pos - station, axis=-1)

        return dist

    def _get_delta_nev_vectors(self):
        temp = self.off_nevs - self.ref_nevs

        # Center to Center distance between reference survey station and
        # closest point in the offset well
        self.distance_cc = norm(temp, axis=-1).reshape(-1, 1)

        with np.errstate(divide='ignore', invalid='ignore'):
            self.ref_delta_nevs = np.nan_to_num(
                temp / self.distance_cc,
                posinf=0.0,
                neginf=0.0
            )

        temp = self.ref_nevs - self.off_nevs
        with np.errstate(divide='ignore', invalid='ignore'):
            self.off_delta_nevs = np.nan_to_num(
                temp / self.distance_cc,
                posinf=0.0,
                neginf=0.0
            )

        self.hoz_bearing = np.arctan2(
            self.ref_delta_nevs[:, 1], self.ref_delta_nevs[:, 0]
        )
        self.hoz_bearing = (
            (np.around(self.hoz_bearing, 6) + np.pi * 2) % (np.pi * 2)
        )

        self.hoz_bearing_deg = (np.degrees(self.hoz_bearing) + 360) % 360

        self._get_delta_hla_vectors()

        self.toolface_bearing = np.arctan2(
            self.ref_delta_hlas[:, 1], self.ref_delta_hlas[:, 0]
        )
        self.toolface_bearing = (
            (np.around(self.toolface_bearing, 6) + np.pi * 2) % (np.pi * 2)
        )

        self.toolface_bearing_deg = (
            np.degrees(self.toolface_bearing) + 360
        ) % 360

        self._traveling_cylinder()

        self.distance_cc = self.distance_cc.reshape(-1)

    def _get_delta_hla_vectors(self):
        self.ref_delta_hlas = np.vstack([
            NEV_to_HLA(s, nev, cov=False)
            for s, nev in zip(
                self.ref.survey_rad, self.ref_delta_nevs
            )
        ])
        self.off_delta_hlas = np.vstack([
            NEV_to_HLA(s, nev, cov=False)
            for s, nev in zip(
                self.off.survey_rad, self.off_delta_nevs
            )
        ])

    def _get_covs(self):
        self.ref_cov_hla = self.ref.cov_hla
        self.ref_cov_nev = self.ref.cov_nev
        self.off_cov_hla = self.off.cov_hla
        self.off_cov_nev = self.off.cov_nev

    def _get_PCRs(self):
        self.ref_pcr = np.hstack([
            np.sqrt(np.dot(np.dot(vec, cov), vec.T))
            for vec, cov in zip(self.ref_delta_nevs, self.ref_cov_nev)
        ])
        self.off_pcr = np.hstack([
            np.sqrt(np.dot(np.dot(vec, cov), vec.T))
            for vec, cov in zip(self.off_delta_nevs, self.off_cov_nev)
        ])

    def _get_calc_hole(self):
        self.calc_hole = self.Rr + self.Ro[self.idx]

    def _traveling_cylinder(self):
        """
        Calculates the azimuthal data for a traveling cylinder plot.
        """
        self.trav_cyl_azi_deg = (
            self.reference.azi_grid_deg[self.kop_index:]
            + self.toolface_bearing_deg + 360
        ) % 360


class MeshClearance(Clearance):
    """
    Class to calculate the clearance between two well bores using a novel
    mesh clearance method. This method is experimental and was developed
    to provide a fast method for determining if well bores are potentially
    colliding.

    This class requires that `trimesh` is installed along with
    `python-fcl`.

    Parameters
    ----------
    n_verts : int
        The number of points (vertices) used to generate the uncertainty
        ellipses which are used to generate a `trimesh` representation of
        the well bores. The default is 12 which is a good balance between
        accuracy and speed.
    sigma : float
        The required/desired sigma value representation of the generated
        mesh. The default value of 2.445 represents about 98.5% confidence
        of the well bore being located within the volume of the generated
        mesh.
    """
    def __init__(
        self,
        *clearance_args,
        n_verts: int = 12,
        sigma: float = 2.445,
        return_data: bool = True,
        return_meshes: bool = False,
        **clearance_kwargs
    ):
        super().__init__(*clearance_args, **clearance_kwargs)

        assert MESH_MODE, "ImportError: try pip install welleng[all]"
        self.n_verts = n_verts
        self.sigma = sigma
        self.Rr = self.ref.radius
        self.Ro = self.offset.radius

        # if you're only interesting in a binary "go/no-go" decision
        # then you can forfeit the expensive ISCWSA calculations by
        # setting return_data to False.
        self.return_data = return_data
        self.collision = []

        if self.return_data:
            self.distance_cc = []
            self.distance = []
            self.off_index = []
            self.sf = []
            self.nev = []
            self.hoz_bearing_deg = []
            self.ref_pcr = []
            self.off_pcr = []
            self.calc_hole = []
            self.ref_md = []
            self.off_md = []

        self.return_meshes = return_meshes
        if self.return_meshes:
            self.meshes = []

        # generate mesh for offset well
        self.off_mesh = self._get_mesh(self.offset, offset=True).mesh

        # make a CollisionManager object and add the offset well mesh
        self.cm = trimesh.collision.CollisionManager()

        self.cm.add_object("offset", self.off_mesh)

        self._process_well()

    def get_lines(self):
        """
        Extracts the closest points between wells for each survey section.
        """
        points = [
            list(d[2]._points.values())
            for d in self.distance
        ]

        start_points = []
        end_points = []

        [(start_points.append(p[0]), end_points.append(p[1])) for p in points]

        return np.array([
            np.vstack(start_points).tolist(),
            np.vstack(end_points).tolist()
        ])

    def _get_mesh(self, survey, offset=False):
        """
        Generates a mesh object from the survey object.
        """
        if offset:
            Sm = self.Sm
        else:
            Sm = 0.0

        mesh = WellMesh(
            survey=survey,
            n_verts=self.n_verts,
            sigma=self.sigma,
            Sm=Sm,
        )

        return mesh

    def _process_well(self):
        """
        Iterates through the reference well survey, determining for
        each section whether a collision has occurred with the offset
        well and optionally calculates separation data.
        """
        ref = self.ref
        off = self.offset
        off_nevs = self.offset_nevs

        for i in range(len(ref.md) - 1):
            # slice a well section and create section survey
            s = slice_survey(ref, i, i + 2)

            # generate a mesh for the section slice
            m = self._get_mesh(s).mesh

            # see if there's a collision
            collision = self.cm.in_collision_single(
                m, return_names=self.return_data, return_data=self.return_data
            )

            if self.return_data:
                distance = self.cm.min_distance_single(
                    m, return_name=True, return_data=True
                )
                closest_point_reference = distance[2].point("__external")
                name_offset_absolute = distance[1]
                closest_point_offset = distance[2].point(name_offset_absolute)

                ref_nev = self._get_closest_nev(s, closest_point_reference)
                ref_md = ref.md[i] + ref_nev[1].x[0]

                # find the closest point on the well trajectory to the closest
                # points on the mesh surface
                off_index = KDTree(off_nevs).query(closest_point_offset)[1]
                if off_index < len(off.md) - 1:
                    s = slice_survey(off, off_index)
                    off_nev_1 = self._get_closest_nev(s, closest_point_offset)
                else:
                    off_nev_1 = False

                if off_index > 0:
                    s = slice_survey(off, off_index - 1)
                    off_nev_0 = self._get_closest_nev(s, closest_point_offset)
                else:
                    off_nev_0 = False

                if off_nev_0 and off_nev_1:
                    if off_nev_0[1].fun < off_nev_1[1].fun:
                        off_nev = off_nev_0
                        off_md = off.md[off_index-1] + off_nev_0[1].x[0]
                    else:
                        off_nev = off_nev_1
                        off_md = off.md[off_index] + off_nev_1[1].x[0]
                elif off_nev_0:
                    off_nev = off_nev_0
                    off_md = off.md[off_index-1] + off_nev_0[1].x[0]
                else:
                    off_nev = off_nev_1
                    off_md = off.md[off_index] + off_nev_1[1].x[0]

                vec = off_nev[0] - ref_nev[0]
                distance_cc = norm(vec)
                hoz_bearing_deg = (
                    np.degrees(np.arctan2(vec[1], vec[0])) + 360
                ) % 360

                if collision[0] is True:
                    depth = norm(
                        closest_point_offset - closest_point_reference
                    )
                    # prevent divide by zero
                    if distance_cc != 0 and depth != 0:
                        sf = distance_cc / (distance_cc + depth)
                    else:
                        sf = 0
                else:
                    sf = distance_cc / (distance_cc - distance[0])

                # data for ISCWSA method comparison
                # self.collision.append(collision)
                self.off_index.append(off_index)
                self.distance.append(distance)
                self.distance_cc.append(distance_cc)
                self.sf.append(round(sf, 2))
                self.nev.append((ref_nev, off_nev))
                self.hoz_bearing_deg.append(hoz_bearing_deg)
                self.ref_pcr.append(
                    (ref_nev[1].fun - self.sigma_pa / 2 - self.Rr[i])
                    / self.sigma
                )
                self.off_pcr.append(
                    (
                        off_nev[1].fun - self.sigma_pa / 2
                        - self.Ro[off_index] - self.Sm
                    ) / self.sigma
                )
                self.calc_hole.append(ref.radius[i] + off.radius[off_index])
                self.ref_md.append(ref_md)
                self.off_md.append(off_md)

            if self.return_meshes:
                self.meshes.append(m)

            # else:
            self.collision.append(collision)

    def _fun(self, x, survey, pos):
        """
        Interpolates a point on a well trajectory and returns
        the distance between the interpolated point and the
        position provided.
        """
        s = _interpolate_survey(survey, x[0])
        new_pos = np.array([s.n, s.e, s.tvd]).T[1]
        dist = norm(new_pos - pos, axis=-1)

        return dist

    def _get_closest_nev(self, survey, pos):
        """
        Using an optimization function to determine the closest
        point along a well trajectory to the position provided.
        """
        bnds = [(0, survey.md[1] - survey.md[0])]
        res = optimize.minimize(
            self._fun,
            bnds[0][1] / 2,
            # method='SLSQP',
            method='Powell',
            bounds=bnds,
            args=(survey, pos)
            )

        s = _interpolate_survey(survey, res.x[0])

        nev = np.array([s.n, s.e, s.tvd]).T[-1]

        return (nev, res)


def get_ref_sigma(sigma1, sigma2, sigma3, kop_index):
    sigma = np.array([sigma1, sigma2, sigma3]).T
    sigma_diff = np.diff(sigma, axis=0)

    sigma_above = np.cumsum(sigma_diff[:kop_index][::-1], axis=0)[::-1]
    sigma_below = np.cumsum(sigma_diff[kop_index:], axis=0)

    sigma_new = np.vstack((sigma_above, np.array([0, 0, 0]), sigma_below))

    return sigma_new



import numpy as np
import math
from copy import copy
try:
    from magnetic_field_calculator import MagneticFieldCalculator
    MAG_CALC = True
except ImportError:
    MAG_CALC = False
from datetime import datetime
from scipy.optimize import minimize
from scipy.spatial.transform import Rotation as R

from .version import __version__
from .utils import (
    MinCurve,
    get_nev,
    get_vec,
    get_angles,
    HLA_to_NEV,
    NEV_to_HLA,
    get_xyz,
    radius_from_dls,
    get_arc
)
from .error import ErrorModel, ERROR_MODELS
from .node import Node
from .connector import Connector, interpolate_well
from .visual import figure
from .units import ureg

from typing import List, Union
from numpy.typing import NDArray


AZI_REF = ["true", "magnetic", "grid"]


class SurveyHeader:
    """
    A class for storing header information about a well.

    Parameters
    ----------
    name: string (default: None)
        The assigned name of the well bore.
    longitude: float (default: None)
        The longitude of the surface location of the well. If left
        default (None) then it will be assigned to Grenwich, the
        undisputed center of the universe.
    latitude: float (default: None)
        The latitude of the surface location of the well. If left
        default (None) then it will be assigned to Grenwich, the
        undisputed center of the universe.
    altitude: float (default: None)
        The altitude of the surface location. If left defaults (None)
        then it will be assigned to 0.
    survey_date: YYYY-mm-dd (default: None)
        The date on which the survey data was recorded. If left
        default then the current date is assigned.
    G: float (default: 9.80665)
        The gravitational field strength in m/s^2.
    b_total: float (default: None)
        The gravitation field strength in nT. If left default, then
        the value is calculated from the longitude, latitude, altitude
        and survey_data properties using the magnetic_field_calculator.
    earth_rate: float (default: 0.26249751949994715)
        The rate of rotation of the earth in radians per hour.
    noise_reduction_factor: float (default: 1.0)
        A fiddle factor for random gyro noise.
    dip: float (default: None)
        The dip (inclination) of the magnetic field relative to the
        earth's horizontal. If left default, then the value is
        calculated using the magnetic_field_calculator. The unit (deg
        of rad) is determined by the deg property.
    declination: float (default: None)
        The angle between true north and magnetic north at the well
        location. If left default, then the value is calculated
        using the magnetic_field_calculator.
    convergence: float (default: 0)
        The angle of convergence between the projection meridian and
        the line from true north through the location of the well.
    azi_reference: string (default: 'true')
        The reference system for the azimuth angles in the survey data,
        either "true", "magnetic" or "grid". Note that survey
        calculations are performed in the "grid" reference and
        converted to and from the other systems.
    vertical_inc_limit: float (default 0.0001)
        For survey inclination angles less than the vertical_inc_limit
        (in degrees), calculations are approximated to avoid
        singularities and errors.
    deg: bool (default: True)
        Indicates whether the survey angles are measured in degrees
        (True) or radians (False).
    depth_unit: string (default: "meters")
        The unit of depth for the survey data, either "meters" or
        "feet".
    surface_unit: string (default: "feet")
        The unit of distance for the survey data, either "meters" or
        "feet".
    vertical_section_azimuth: float (default: 0.0)
        The azimuth along which to determine the vertical section data
        for the well trajectory.
    """
    def __init__(
        self,
        name: str = None,
        longitude=None,
        latitude=None,
        altitude=None,
        survey_date=None,
        G=9.80665,
        b_total=None,
        earth_rate=0.26251614,
        dip=None,
        declination=None,
        convergence=0,
        azi_reference="true",
        vertical_inc_limit=0.0001,
        deg=True,
        depth_unit='meters',
        surface_unit='meters',
        mag_defaults={
            'b_total': 50_000.,
            'dip': 70.,
            'declination': 0.,
        },
        vertical_section_azimuth=0,
        # **kwargs
    ):
        if latitude is not None:
            assert 90 >= latitude >= -90, "latitude out of bounds"
        if longitude is not None:
            assert 180 >= longitude >= -180, "longitude out of bounds"
        assert azi_reference in AZI_REF

        self._validate_date(survey_date)
        self.name = name
        self.latitude = latitude if latitude is not None else 51.4934
        self.longitude = longitude if longitude is not None else 0.0098
        self.altitude = altitude if altitude is not None else 0.
        self._get_date(survey_date)
        self.b_total = b_total
        self.earth_rate = earth_rate
        self.dip = dip
        self.convergence = convergence
        self.declination = declination
        self.vertical_inc_limit = vertical_inc_limit

        self.depth_unit = get_unit(depth_unit)
        self.surface_unit = get_unit(surface_unit)
        self.G = G
        self.azi_reference = azi_reference
        self.vertical_section_azimuth = vertical_section_azimuth

        self.mag_defaults = mag_defaults
        self._get_mag_data(deg)

    def _get_mag_data(self, deg):
        """
        Initiates b_total if provided, else calculates a value.
        """
        if MAG_CALC:
            calculator = MagneticFieldCalculator()
            try:
                result = calculator.calculate(
                    latitude=self.latitude,
                    longitude=self.longitude,
                    altitude=self.altitude,
                    date=self.survey_date
                )
            except:
                result = calculator.calculate(
                    latitude=self.latitude,
                    longitude=self.longitude,
                    altitude=self.altitude,
                    date=self._get_date(date=None)
                )
        else:
            result = {
                'field-value': {
                    'total-intensity': {
                        'value': self.mag_defaults.get('b_total')
                    },
                    'inclination': {
                        'value': self.mag_defaults.get('dip')
                    },
                    'declination': {
                        'value': self.mag_defaults.get('declination')
                    }
                }
            }

        if self.b_total is None:
            self.b_total = result['field-value']['total-intensity']['value']
            # if not deg:
            #     self.b_total = math.radians(self.b_total)
        if self.dip is None:
            self.dip = -result['field-value']['inclination']['value']
            if not deg:
                self.dip = math.radians(self.dip)
        if self.declination is None:
            self.declination = result['field-value']['declination']['value']
            if not deg:
                self.declination = math.radians(self.declination)

        if deg:
            self.dip = math.radians(self.dip)
            self.declination = math.radians(self.declination)
            self.convergence = math.radians(self.convergence)
            self.vertical_inc_limit = math.radians(
                self.vertical_inc_limit
            )
            self.vertical_section_azimuth = math.radians(
                self.vertical_section_azimuth
            )

    def _get_date(self, date):
        if date is None:
            date = datetime.today().strftime('%Y-%m-%d')
        self.survey_date = date

    def _validate_date(self, date):
        if date is None:
            return
        try:
            datetime.strptime(date, '%Y-%m-%d')
        except ValueError:
            raise ValueError("incorrect data format, should be YYYY-MM-DD")


class Survey:
    """
    Initialize a `welleng.Survey` object. Calculations are performed in the
    azi_reference "grid" domain.

    Parameters
    ----------
    md: (,n) list or array of floats
        List or array of well bore measured depths.
    inc: (,n) list or array of floats
        List or array of well bore survey inclinations
    azi: (,n) list or array of floats
        List or array of well bore survey azimuths
    n: (,n) list or array of floats (default: None)
        List or array of well bore northings
    e: (,n) list or array of floats (default: None)
        List or array of well bore eastings
    tvd: (,n) list or array of floats (default: None)
        List or array of local well bore z coordinates, i.e. depth
        and usually relative to surface or mean sea level.
    x: (,n) list or array of floats (default: None)
        List or array of local well bore x coordinates, which is
        usually aligned to the east direction.
    y: (,n) list or array of floats (default: None)
        List or array of local well bore y coordinates, which is
        usually aligned to the north direction.
    z: (,n) list or array of floats (default: None)
        List or array of well bore true vertical depths relative
        to the well surface datum (usually the drill floor
        elevation DFE, so not always identical to tvd).
    vec: (n,3) list or array of (,3) floats (default: None)
        List or array of well bore unit vectors that describe the
        inclination and azimuth of the well relative to (x,y,z)
        coordinates.
    header: SurveyHeader object (default: None)
        A SurveyHeader object with information about the well location
        and survey data. If left default then a SurveyHeader will be
        generated with the default properties assigned, but these may
        not be relevant and may result in incorrect data.
    radius: float or (,n) list or array of floats (default: None)
        If a single float is specified, this value will be
        assigned to the entire well bore. If a list or array of
        floats is provided, these are the radii of the well bore.
        If None, a well bore radius of 12" or approximately 0.3 m
        is applied.
    cov_nev: (n,3,3) list or array of floats (default: None)
        List or array of covariance matrices in the (n,e,v)
        coordinate system.
    cov_hla: (n,3,3) list or array of floats (default: None)
        List or array of covariance matrices in the (h,l,a)
        well bore coordinate system (high side, lateral, along
        hole).
    error_model: str (default: None)
        If specified, this model is used to calculate the
        covariance matrices if they are not present. Currently,
        only the "ISCWSA_MWD" model is provided.
    start_xyz: (,3) list or array of floats (default: [0,0,0])
        The start position of the well bore in (x,y,z) coordinates.
    start_nev: (,3) list or array of floats (default: [0,0,0])
        The start position of the well bore in (n,e,v) coordinates.
    start_cov_nev: (,3,3) list or array of floats (default: None)
        The covariance matrix for the start position of the well
        bore in (n,e,v) coordinates.
    deg: boolean (default: True)
        Indicates whether the provided angles are in degrees
        (True), else radians (False).
    unit: str (default: 'meters')
        Indicates whether the provided lengths and distances are
        in 'meters' or 'feet', which impacts the calculation of
        the dls (dog leg severity).

    Returns
    -------
    A welleng.survey.Survey object.
    """
    def __init__(
        self,
        md,
        inc,
        azi,
        n=None,
        e=None,
        tvd=None,
        x=None,
        y=None,
        z=None,
        vec=None,
        nev=True,
        header=None,
        radius=None,
        cov_nev=None,
        cov_hla=None,
        error_model=None,
        start_xyz=[0., 0., 0.],
        start_nev=[0., 0., 0.],
        start_cov_nev=None,
        deg=True,
        unit="meters",
        **kwargs
    ):
        if header is None:
            self.header = SurveyHeader()
        else:
            assert isinstance(header, SurveyHeader)
            self.header = header
        assert unit == self.header.depth_unit, (
            "inconsistent units with header"
        )

        self.azi_ref_lookup = {
            'true': "true", 'magnetic': "mag", 'grid': "grid"
        }

        self.unit = unit
        self.deg = deg
        self.start_xyz = start_xyz
        self.start_nev = start_nev
        self.md = np.array(md).astype('float64')
        self.start_cov_nev = start_cov_nev

        self._process_azi_ref(inc, azi, deg)

        self._get_radius(radius)

        self.survey_deg = np.array(
            [self.md, self.inc_deg, self.azi_grid_deg]
        ).T
        self.survey_rad = np.array(
            [self.md, self.inc_rad, self.azi_grid_rad]
        ).T

        self.n = np.array(n) if n is not None else n
        self.e = np.array(e) if e is not None else e
        self.tvd = np.array(tvd) if tvd is not None else tvd

        # start_nev will be overwritten if n, e, tvd data provided
        if not all((self.n is None, self.e is None, self.tvd is None)):
            self.start_nev = np.array(
                [self.n[0], self.e[0], self.tvd[0]]
            )
        else:
            self.start_nev = np.array(start_nev)

        self.x = np.array(x) if x is not None else x
        self.y = np.array(y) if y is not None else y
        self.z = np.array(z) if z is not None else z
        if vec is not None:
            if nev:
                self.vec_nev = vec
                self.vec_xyz = get_xyz(vec)
            else:
                self.vec_xyz = vec
                self.vec_nev = get_nev(vec)
        else:
            self.vec_nev, self.vec_xyz = vec, vec

        self._min_curve(vec)
        self._get_toolface_and_rates()

        # initialize errors
        # TODO: read this from a yaml file in errors
        error_models = ERROR_MODELS
        if error_model is not None:
            assert error_model in error_models, "Unrecognized error model"
        self.error_model = error_model

        self.cov_hla = cov_hla
        self.cov_nev = cov_nev

        self._get_errors()

        self.interpolated = (
            np.full_like(self.md, False)
            if kwargs.get('interpolated') is None
            else kwargs.get('interpolated')
        )

        self._get_vertical_section()

    def _process_azi_ref(self, inc, azi, deg):
        if self.header.azi_reference == 'grid':
            self._make_angles(inc, azi, deg)
            self.azi_true_deg = (
                self.azi_grid_deg + math.degrees(self.header.convergence)
            )
            self.azi_mag_deg = (
                self.azi_true_deg - math.degrees(self.header.declination)
            )
            self._get_azi_mag_and_true_rad()
        elif self.header.azi_reference == 'true':
            if deg:
                self.azi_true_deg = np.array(azi).astype('float64')
                self.azi_mag_deg = (
                    self.azi_true_deg - math.degrees(self.header.declination)
                )
                self._get_azi_mag_and_true_rad()
                azi_temp = self._get_azi_temp(deg)
            else:
                self.azi_true_rad = np.array(azi).astype('float64')
                self.azi_mag_rad = (
                    self.azi_true_rad - self.header.declination
                )
                self._get_azi_mag_and_true_deg()
                azi_temp = self._get_azi_temp(deg)
            self._make_angles(inc, azi_temp, deg)
        else:  # azi_reference is "magnetic"
            if deg:
                self.azi_mag_deg = np.array(azi).astype('float64')
                self.azi_true_deg = (
                    self.azi_mag_deg + math.degrees(self.header.declination)
                )
                self._get_azi_mag_and_true_rad()
                azi_temp = self._get_azi_temp(deg)
            else:
                self.azi_mag_rad = np.array(azi).astype('float64')
                self.azi_true_rad = (
                    self.azi_mag_rad + self.header.declination
                )
                self._get_azi_mag_and_true_deg()
                azi_temp = self._get_azi_temp(deg)
            self._make_angles(inc, azi_temp, deg)

    def _get_azi_temp(self, deg):
        if deg:
            azi_temp = self.azi_true_deg - math.degrees(
                self.header.convergence
            )
        else:
            azi_temp = self.azi_true_rad - self.header.convergence

        return azi_temp

    def _get_azi_mag_and_true_rad(self):
        self.azi_true_rad, self.azi_mag_rad = (
            np.radians(np.array([
                self.azi_true_deg, self.azi_mag_deg
            ]))
        )

    def _get_azi_mag_and_true_deg(self):
        self.azi_true_deg, self.azi_mag_deg = (
            np.degrees(np.array([
                self.azi_true_rad, self.azi_mag_rad
            ]))
        )

    def _get_radius(self, radius=None):
        if radius is None:
            self.radius = np.full_like(self.md.astype(float), 0.3048)
        elif np.array([radius]).shape[-1] == 1:
            self.radius = np.full_like(self.md.astype(float), radius)
        else:
            assert len(radius) == len(self.md), "Check radius"
            self.radius = np.array(radius)

    def _min_curve(self, vec):
        """
        Get the (x,y,z), (n,e,v), doglegs, rfs, delta_mds, dlss and
        vectors for the well bore if they were not provided, using the
        minimum curvature method.
        """
        mc = MinCurve(
            self.md, self.inc_rad, self.azi_grid_rad, self.start_xyz, self.unit
        )
        self.dogleg = mc.dogleg
        self.rf = mc.rf
        self.delta_md = mc.delta_md
        self.dls = mc.dls
        self.pos_xyz = mc.poss
        self.pos_nev = get_nev(self.pos_xyz) + self.start_nev

        if self.x is None:
            # self.x, self.y, self.z = (mc.poss + self.start_xyz).T
            self.x, self.y, self.z = (mc.poss).T
        if self.n is None:
            self._get_nev()
        if vec is None:
            self.vec_xyz = get_vec(self.inc_rad, self.azi_grid_rad, deg=False)
            self.vec_nev = get_vec(
                self.inc_rad, self.azi_grid_rad, deg=False, nev=True
            )

    def _get_nev(self):
        self.n, self.e, self.tvd = get_nev(
            np.array([
                self.x,
                self.y,
                self.z
            ]).T,
            start_xyz=self.start_xyz,
            start_nev=self.start_nev
        ).reshape(-1, 3).T

    def _make_angles(self, inc, azi, deg=True):
        """
        Calculate angles in radians if they were provided in degrees or
        vice versa.
        """
        if deg:
            self.inc_rad = np.radians(inc)
            self.azi_grid_rad = np.radians(azi)
            self.inc_deg = np.array(inc)
            self.azi_grid_deg = np.array(azi)
        else:
            self.inc_rad = np.array(inc)
            self.azi_grid_rad = np.array(azi)
            self.inc_deg = np.degrees(inc)
            self.azi_grid_deg = np.degrees(azi)

    def get_error(self, error_model, return_error=False):
        assert error_model in ERROR_MODELS, "Undefined error model"

        self.error_model = error_model
        self._get_errors()

        if return_error:
            return self.err
        else:
            return self

    def _get_errors(self):
        """
        Initiate a welleng.error.ErrorModel object and calculate the
        covariance matrices with the specified error model.
        """
        if self.error_model:
            # if self.error_model == "iscwsa_mwd_rev4":
            self.err = ErrorModel(
                self,
                error_model=self.error_model
            )
            self.cov_hla = self.err.errors.cov_HLAs.T
            self.cov_nev = self.err.errors.cov_NEVs.T
        else:
            if self.cov_nev is not None and self.cov_hla is None:
                self.cov_hla = NEV_to_HLA(self.survey_rad, self.cov_nev.T).T
            elif self.cov_nev is None and self.cov_hla is not None:
                self.cov_nev = HLA_to_NEV(self.survey_rad, self.cov_hla.T).T
            else:
                pass

        if (
            self.start_cov_nev is not None
            and self.cov_nev is not None
        ):
            self.cov_nev += self.start_cov_nev
            self.cov_hla = NEV_to_HLA(self.survey_rad, self.cov_nev.T).T

    def _curvature_to_rate(self, curvature):
        with np.errstate(divide='ignore', invalid='ignore'):
            radius = 1 / curvature
        circumference = 2 * np.pi * radius
        if self.unit == 'meters':
            x = 30
        else:
            x = 100
        rate = np.absolute(np.degrees(2 * np.pi / circumference) * x)

        return rate

    def _get_toolface_and_rates(self):
        """
        Reference SPE-84246.
        theta is inc, phi is azi
        """
        # split the survey
        s = SplitSurvey(self)

        if self.unit == 'meters':
            x = 30
        else:
            x = 100

        # this is lazy I know, but I'm using this mostly for flags
        with np.errstate(divide='ignore', invalid='ignore'):
            t1 = np.arctan2(
                np.sin(s.inc2) * np.sin(s.delta_azi),
                (
                    np.sin(s.inc2) * np.cos(s.inc1) * np.cos(s.delta_azi)
                    - np.sin(s.inc1) * np.cos(s.inc2)
                )
            )
            t1 = np.nan_to_num(
                t1,
                # np.where(t1 < 0, t1 + 2 * np.pi, t1),
                nan=np.nan
            )
            t2 = np.arctan2(
                np.sin(s.inc1) * np.sin(s.delta_azi),
                (
                    np.sin(s.inc2) * np.cos(s.inc1)
                    - np.sin(s.inc1) * np.cos(s.inc2) * np.cos(s.delta_azi)
                )
            )
            t2 = np.nan_to_num(
                np.where(t2 < 0, t2 + 2 * np.pi, t2),
                nan=np.nan
            )
            self.curve_radius = (360 / self.dls * x) / (2 * np.pi)

            curvature_dls = 1 / self.curve_radius

            self.toolface = np.concatenate((t1, np.array([t2[-1]])))

            curvature_turn = curvature_dls * (
                np.sin(self.toolface) / np.sin(self.inc_rad)
            )
            self.turn_rate = self._curvature_to_rate(curvature_turn)

            curvature_build = curvature_dls * np.cos(self.toolface)
            self.build_rate = self._curvature_to_rate(curvature_build)

        # calculate plane normals
        # TODO: update so its the same length as the survey - need to cascade
        n12 = np.cross(s.vec1_nev, s.vec2_nev)
        with np.errstate(divide='ignore', invalid='ignore'):
            self.normals = n12 / np.linalg.norm(n12, axis=1).reshape(-1, 1)

        # get radius vectors
        with np.errstate(divide='ignore', invalid='ignore'):
            self.vec_radius_nev = np.cross(
                np.vstack((self.normals[0], self.normals)),  # temp fix
                self.vec_nev
            )

    def _get_sections(self, rtol=0.1, atol=0.1, dls_cont=True):
        sections = get_sections(self, rtol, atol, dls_cont)

        return sections

    def get_nev_arr(self):
        return np.array([
            self.n,
            self.e,
            self.tvd
        ]).T.reshape(-1, 3)

    def save(self, filename):
        """
        Saves a minimal (control points) survey listing as a .csv file,
        including the survey header information.

        Parameters
        ----------
        filename: str
            The path and filename for saving the text file.
        """
        export_csv(self, filename)

    def interpolate_md(self, md):
        """
        Method to interpolate a position based on measured depth and return
        a node.

        Parameters
        ----------
        md: float
            The measured depth of the point of interest.

        Returns
        -------
        node: we.node.Node object
            A node with attributes describing the point at the provided
            measured depth.

        Examples
        --------
        >>> import welleng as we
        >>> survey = we.connector.interpolate_survey(
        ...    survey=we.survey.Survey(
        ...       md=[0, 500, 1000, 2000, 3000],
        ...       inc=[0, 0, 30, 90, 90],
        ...       azi=[0, 0, 45, 135, 180],
        ...    ),
        ...    step=30
        ... )
        >>> node = survey.interpolate_md(1234)
        >>> node.properties()
        {
            'vec_nev': [0.07584209568113438, 0.5840332282889957, 0.8081789187902809],
            'vec_xyz': [0.5840332282889957, 0.07584209568113438, 0.8081789187902809],
            'inc_rad': 0.6297429542197106,
            'azi_rad': 1.4416597719915565,
            'inc_deg': 36.081613454889634,
            'azi_deg': 82.60102042890875,
            'pos_nev': [141.27728744087796, 201.41424652428694, 1175.5823295305202],
            'pos_xyz': [201.41424652428694, 141.27728744087796, 1175.5823295305202],
            'md': 1234.0,
            'unit': 'meters',
            'interpolated': True
        }
        """
        s = interpolate_md(self, md)
        node = get_node(s, -1, s.interpolated[-1])

        return node

    def interpolate_tvd(self, tvd):
        node = interpolate_tvd(self, tvd=tvd)
        return node

    def interpolate_survey_tvd(self, start=None, stop=None, step=10):
        """
        Convenience method for interpolating a Survey object's TVD.
        """
        survey_interpolated = interpolate_survey_tvd(
            self, start=start, stop=stop, step=step
        )
        return survey_interpolated

    def interpolate_survey(self, step=30, dls=1e-8):
        """
        Convenience method for interpolating a Survey object's MD.
        """
        survey_interpolated = interpolate_survey(self, step=step, dls=dls)
        return survey_interpolated

    def figure(self, type='scatter3d', **kwargs):
        fig = figure(self, type, **kwargs)
        return fig

    def project_to_bit(self, delta_md, dls=None, toolface=None):
        """
        Convenience method to project the survey ahead to the bit.

        Parameters
        ----------
        delta_md: float
            The along hole distance from the surveying tool to the bit in
            meters.
        dls: float
            The desired dog leg severity (deg / 30m) between the surveying
            tool and the bit. Default is to project the DLS of the last
            survey section.
        toolface: float
            The desired toolface to project from at the last survey point.
            The default is to project the current toolface from the last
            survey station.

        Returns
        -------
        node: welleng.node.Node object
        """
        if dls is None:
            dls = self.dls[-1]
        if toolface is None:
            toolface = self.toolface[-1]

        node = project_ahead(
            pos=np.array([self.n, self.e, self.tvd]).T[-1],
            vec=self.vec_nev[-1],
            delta_md=delta_md,
            dls=dls,
            toolface=toolface,
            md=self.md[-1]
        )

        return node

    def project_to_target(
        self,
        node_target,
        dls_design=3.0,
        delta_md=None,
        dls=None, toolface=None,
        step=30
    ):
        survey = project_to_target(
            self,
            node_target,
            dls_design,
            delta_md,
            dls, toolface,
            step
        )
        return survey

    def _get_vertical_section(self, *args, **kwargs):
        """
        Internal function to initiate the vertical section by calculating
        the magnitude of the lateral displacement and, if a vertical section
        azimuth is defined, calculating the vertical section lateral
        component along the given vertical section azimuth (relative to the
        defined reference azimuth).
        """
        self.hypot = np.hypot(
            self.n[1:] - self.n[:-1],
            self.e[1:] - self.e[:-1]
        )

        if self.header.vertical_section_azimuth is not None:
            self.vertical_section = self.get_vertical_section(
                self.header.vertical_section_azimuth, deg=False
            )
        else:
            self.vertical_section = None

    def get_vertical_section(self, vertical_section_azimuth, deg=True):
        """
        Calculate the vertical section.

        Parameters
        ----------
        vertical_section_azimuth: float
            The azimuth (relative to the reference azimuth defined in the
            survey header) along which to calculate the vertical section
            lateral displacement.
        deg: boolean (default: True)
            Indicates whether the vertical section azimuth parameter is in
            degrees or radians (True or False respectively).

        Returns
        -------
        result: (n, 1) ndarray
        """
        vertical_section_azimuth = (
            vertical_section_azimuth if not deg
            else math.radians(vertical_section_azimuth)
        )
        azi_temp = getattr(
            self,
            f"azi_{self.azi_ref_lookup[self.header.azi_reference]}_rad"
        )
        azi_temp[np.where(self.inc_rad == 0.0)] = (
            vertical_section_azimuth
        )
        result = np.cos(
            vertical_section_azimuth
            - (azi_temp[1:] + azi_temp[:-1]) / 2
        ) * self.hypot

        result = np.cumsum(np.hstack(([0.0], result)))

        return result

    def set_vertical_section(self, vertical_section_azimuth, deg=True):
        """
        Sets the vertical_section_azimuth property in the survey header and
        the vertical section data with the data calculated for the input
        azimuth.

        Parameters
        ----------
        vertical_section_azimuth: float
            The azimuth (relative to the reference azimuth defined in the
            survey header) along which to calculate the vertical section
            lateral displacement.
        deg: boolean (default: True)
            Indicates whether the vertical section azimuth parameter is in
            degrees or radians (True or False respectively).
        """
        self.header.vertical_section_azimuth = vertical_section_azimuth
        self.vertical_section = self.get_vertical_section(
            vertical_section_azimuth, deg
        )

    def modified_tortuosity_index(
        self, rtol=1.0, dls_tol=1e-3, step=1.0, dls_noise=1.0, data=False,
        **kwargs
    ):
        """
        Convenience method for calculating the Tortuosity Index (TI) using a
        modified version of the method described in the [International
        Association of Directional Drilling presentation](https://www.iadd-intl.org/media/files/files/47d68cb4/iadd-luncheon-february-22-2018-v2.pdf)
        by Pradeep Ashok et al.

        Parameters
        ----------
        rtol: float
            Relative tolerance when determining closeness of normal vectors.
        dls_tol: float or None
            Indicates whether or not to check for dls continuity within the
            defined dls tolerance.
        step: float or None
            The step length in meters used for interpolating the survey prior
            to calculating trajectory with the maximum curvature method. If
            None or dls_noise is None then no interpolation is done.
        dls_noise: float or None
            The incremental Dog Leg Severity to be added when using the
            maximum curvature method. If None then no pre-processing will be
            done and minimum curvature is assumed.
        data: bool
            If true returns a dictionary of properties.

        Returns
        -------
        ti: (n,1) array or dict
            Array of tortuosity index or a dict of results where:

        References
        ----------
        Further details regarding the maximum curvature method can be read
        [here](https://jonnymaserati.github.io/2022/06/19/modified-tortuosity-index-survey-frequency.html)
        """
        # Check whether to pre-process the survey to apply maximum curvature.
        if bool(dls_noise):
            survey = self.interpolate_survey(step=step)
            survey = survey.maximum_curvature(dls_noise=dls_noise)

        else:
            survey = self

        return modified_tortuosity_index(
            survey, rtol=rtol, dls_tol=dls_tol, data=data, **kwargs
        )

    def tortuosity_index(self, rtol=0.01, dls_tol=None, data=False, **kwargs):
        """
        A modified version of the Tortuosity Index function originally
        referenced in an IADD presentation on "Measuring Wellbore
        Tortuosity" by Pradeep Ashok - https://www.iadd-intl.org/media/
        files/files/47d68cb4/iadd-luncheon-february-22-2018-v2.pdf with
        reference to the original paper "A Novel Method for the Automatic
        Grading of Retinal Vessel Tortuosity" by Enrico Grisan et al.
        In SPE/IADC-194099-MS there's mention that a factor of 1e7 is
        applied to the TI result since the results are otherwise very small
        numbers.
        Unlike the documented version that uses delta inc and azi for
        determining continuity and directional intervals (which are not
        independent and so values are double dipped in 3D), this method
        determines the point around which the well bore is turning and tests
        for continuity of these points. As such, this function takes account
        of torsion of the well bore and demonstrates that actual/drilled
        trajectories are significantly more tortuous than planned
        trajectories (intuitively).

        Parameters
        ----------
        rtol: float
            Relative tolerance when determining closeness of normal vectors.
        atol: float
            Absolute tolerance when determining closeness of normal vectors.
        data: boolean
            If true returns a dictionary of properties.

        Returns
        ti: (n,1) array
            Array of tortuosity index or a dict or results.
        """

        return tortuosity_index(
            self, rtol=rtol, dls_tol=None, data=data, **kwargs
        )

    def directional_difficulty_index(self, **kwargs):
        """
        Taken from IADC/SPE 59196 The Directional Difficulty Index - A
        New Approach to Performance Benchmarking by Alistair W. Oag et al.

        Returns
        -------
        data: (n) array of floats
            The ddi for each survey station.
        """

        return directional_difficulty_index(self, **kwargs)

    def maximum_curvature(self, dls_noise=1.0):
        """
        Create a well trajectory using the Maximum Curvature method.

        Parameters
        ----------
        survey: welleng.survey.Survey object
        dls_noise: float
            The additional Dog Leg Severity (DLS) in deg/30m used to calculate
            the curvature for the initial section of the survey interval.

        Returns
        -------
        survey_new: welleng.Survey.survey object
            A revised survey object calculated using the Minimum Curvature
            method with updated survey positions and additional mid-point
            stations.
        """

        dls_effective = self.dls + dls_noise
        radius_effective = radius_from_dls(dls_effective)

        dogleg1 = (
            (self.delta_md / radius_effective) / 2
        )

        radius_effective = np.where(
            dogleg1 > np.pi,
            self.delta_md * 4 / (2 * np.pi),
            radius_effective
        )

        arc1 = [
            get_arc(dogleg, _radius_effective, toolface, vec=vec)
            for dogleg, _radius_effective, toolface, vec in zip(
                dogleg1[1:], radius_effective[1:], self.toolface[:-1],
                self.vec_nev[:-1]
            )
        ]

        _survey_new = np.array([
            [row[-1], *np.degrees(get_angles(row[1], nev=True))[0]]
            for row in arc1
        ])
        _survey_new[:, 0] += self.md[:-1]

        survey_new = np.zeros(shape=(len(_survey_new) * 2 + 1, 3))
        survey_new[:-1] = np.stack(
            (self.survey_deg[:-1].T, _survey_new.T),
            axis=1
        ).T.reshape(-1, 3)
        survey_new[-1] = self.survey_deg[-1]

        # Update the new survey header as the new azimuth reference is 'grid'.
        sh = self.header
        sh.azi_reference = 'grid'

        # Create a new Survey instance
        survey = Survey(
            md=survey_new[:, 0],
            inc=survey_new[:, 1],
            azi=survey_new[:, 2],
            header=sh,
            start_xyz=self.start_xyz,
            start_nev=self.start_nev
        )

        # Update the interpolated property to keep track of the original survey
        # stations.
        survey.interpolated = np.full_like(survey.md, True)
        survey.interpolated[::2] = self.interpolated

        return survey


def modified_tortuosity_index(
    survey, rtol=1.0, dls_tol=1e-3, data=False, **kwargs
):
    """
    Method for calculating the Tortuosity Index (TI) using a modified
    version of the method described in the International Association of
    Directional Drilling presentation (https://www.iadd-intl.org/media/files/files/47d68cb4/iadd-luncheon-february-22-2018-v2.pdf)
    by Pradeep Ashok et al.
    """
    # set default params
    coeff = kwargs.get('coeff', 1.0)  # for testing dimensionlessness
    kappa = kwargs.get('kapa', 1)

    continuous, starts, mds, locs, n_sections, n_sections_arr = _get_ti_data(
        survey, rtol, dls_tol
    )

    l_cs = (
        survey.md[1:] - mds[n_sections_arr - 1]
    ) / coeff
    l_xs = np.linalg.norm(
        survey.pos_nev[1:]
        - np.array(locs)[n_sections_arr - 1],
        axis=1
    ) / coeff
    b = (
        (l_cs / l_xs) - 1
    ) / l_cs
    # )

    cumsum = 0
    a = []
    for n in n_sections:
        a.extend(
            b[n_sections_arr == n]
            + cumsum
        )
        cumsum = a[-1]
    a = np.array(a)

    mti = np.hstack((
        np.array([0.0]),
        (
            # 1
            (n_sections_arr / (n_sections_arr + 1))
            * (kappa * ((survey.md[1:] - survey.md[0]) / coeff))
            # * (kappa / (np.cumsum(survey.dogleg)[1:] + 1))
            * a
        )
    ))

    if data:
        return {
            'continuous': continuous, 'starts': starts, 'mds': mds,
            'locs': locs, 'n_sections': n_sections,
            'n_sections_arr': n_sections_arr, 'l_cs': l_cs, 'l_xs': l_xs,
            'mti': mti, 'survey': survey
        }

    return mti


def tortuosity_index(survey, rtol=0.01, dls_tol=None, data=False, **kwargs):
    """
    Method for calculating the Tortuosity Index (TI) as described in the 
    International Association of Directional Drilling presentation
    (https://www.iadd-intl.org/media/files/files/47d68cb4/iadd-luncheon-february-22-2018-v2.pdf)
    by Pradeep Ashok et al.
    """
    # set default params
    coeff = kwargs.get('coeff', 0.3048)  # for testing dimensionlessness
    kappa = kwargs.get('kapa', 1e7)

    continuous, starts, mds, locs, n_sections, n_sections_arr = _get_ti_data(
        survey, rtol, dls_tol
    )

    l_cs = (survey.md[1:] - mds[n_sections_arr - 1]) / coeff
    l_xs = np.linalg.norm(
        survey.pos_nev[1:]
        - np.array(locs)[n_sections_arr - 1],
        axis=1
    ) / coeff
    b = (
        (l_cs / l_xs) - 1
    )

    cumsum = 0
    a = []
    for n in n_sections:
        a.extend(
            b[n_sections_arr == n]
            + cumsum
        )
        cumsum = a[-1]
    a = np.array(a)

    ti = np.hstack((
        np.array([0.0]),
        (
            (n_sections_arr / (n_sections_arr + 1))
            * (kappa / (survey.md[1:] - survey.md[0] / coeff))
            * a
        )
    ))

    if data:
        return {
            'continuous': continuous, 'starts': starts, 'mds': mds,
            'locs': locs, 'n_sections': n_sections,
            'n_sections_arr': n_sections_arr, 'l_cs': l_cs, 'l_xs': l_xs,
            'ti': ti
        }

    return ti


def directional_difficulty_index(survey, **kwargs):
    """
    Taken from IADC/SPE 59196 The Directional Difficulty Index - A
    New Approach to Performance Benchmarking by Alistair W. Oag et al.
    Parameters
    ----------
    survey: welleng.survey.Survey object
    data: bool
        If True, returns the ddi at each survey station.
    Returns
    -------
    ddi: float
        The ddi for the well at well (at TD).
    data: (n) array of floats
        The ddi for each survey station.
    """
    with np.errstate(divide='ignore', invalid='ignore'):
        ddi = np.nan_to_num(np.log10(
            (
                (survey.md * ureg.meters).to('ft').m
                * (
                    np.linalg.norm(
                        (survey.n, survey.e), axis=0
                    ) * ureg.meters
                ).to('ft').m
                * np.cumsum(np.degrees(survey.dogleg))
            )
            / (survey.tvd * ureg.meters).to('ft').m
        ), nan=0.0, posinf=0.0, neginf=0.0)

    return ddi


def _get_ti_data(survey, rtol, dls_tol=None):
    # tol_0 = np.full((len(survey.md) - 2, 3), rtol)
    # delta_md = (survey.md[2:] - survey.md[1:-1]).reshape(-1, 1)
    # delta_norm = survey.normals[1:] - survey.normals[:-1]
    # tol_norm = tol_0 * np.sin(delta_md / 30)

    # continuous = (
    #     (delta_norm / 30 * delta_md <= tol_norm).all(axis=-1) | (
    #         np.isnan(survey.normals[1:]).all(axis=-1)
    #         & np.isnan(survey.normals[:-1]).all(axis=-1)
    #     )
    # )
    if dls_tol is None:
        dls_continuity = np.full(len(survey.dls) - 2, True)
    else:
        dls_continuity = np.isclose(
            survey.dls[1:-1],
            survey.dls[2:],
            equal_nan=True,
            rtol=dls_tol,
            atol=rtol
        )
    continuous = np.all((
        np.all(
            np.isclose(
                survey.normals[:-1],
                survey.normals[1:],
                equal_nan=True,
                rtol=rtol, atol=rtol
            ), axis=-1
        ),
        dls_continuity
    ), axis=0)

    starts = np.concatenate((
        np.array([0]),
        np.where(continuous == False)[0] + 1,
    ))

    mds = survey.md[starts]
    locs = survey.pos_nev[starts]
    n_sections = np.arange(1, len(starts) + 1, 1)
    n_sections_arr = np.searchsorted(mds, survey.md[1:])

    return (
        continuous, starts, mds, locs, n_sections, n_sections_arr
    )


class TurnPoint:
    def __init__(
        self,
        md=None,
        inc=None,
        azi=None,
        build_rate=None,
        turn_rate=None,
        dls=None,
        toolface=None,
        method=None,
        target=None,
        tie_on=False,
        location=None
    ):
        self.md = md
        self.inc = inc
        self.azi = azi
        self.build_rate = build_rate
        self.turn_rate = turn_rate
        self.dls = dls
        self.toolface = toolface
        self.method = method
        self.target = target
        self.tie_on = tie_on
        self.location = location


def get_node(survey, idx, interpolated=False):
    node = Node(
        pos=[survey.n[idx], survey.e[idx], survey.tvd[idx]],
        vec=survey.vec_nev[idx].tolist(),
        md=survey.md[idx],
        unit=survey.unit,
        nev=True,
        interpolated=interpolated
    )
    return node


def interpolate_md(survey, md):
    """
    Interpolates a survey at a given measured depth.
    """
    # get the closest survey stations
    idx = np.searchsorted(survey.md, md, side="left") - 1

    assert idx < len(survey.md), "The md is beyond the survey"

    if idx < 0:
        idx = 0
        x = 0

    else:
        x = md - survey.md[idx]

    return _interpolate_survey(survey, x=x, index=idx)


def _interpolate_survey(survey, x=0, index=0):
    """
    Interpolates a point distance x between two survey stations
    using minimum curvature.

    Parameters
    ----------
        survey: welleng.Survey
            A survey object with at least two survey stations.
        x: float
            Length along well path from indexed survey station to
            perform the interpolate at. Must be less than length
            to the next survey station.
        index: int
            The index of the survey station from which to interpolate
            from.

    Returns
    -------
        survey: welleng.Survey
            A survey object consisting of the two survey stations
            between which the interpolation has been made (index 0 and
            2), with the interpolated station between them (index 1)

    """
    index = int(index)

    assert index < len(survey.md) - 1, "Index is out of range"

    # check if it's just a tangent section
    if survey.dogleg[index + 1] == 0:
        azi = survey.azi_grid_rad[index]
        inc = survey.inc_rad[index]

    else:
        # get the vector
        t1 = survey.vec_xyz[index]
        t2 = survey.vec_xyz[index + 1]

        total_dogleg = survey.dogleg[index + 1]

        dogleg = x * (total_dogleg / survey.delta_md[index + 1])

        t = (
            (math.sin(total_dogleg - dogleg) / math.sin(total_dogleg)) * t1
            + (math.sin(dogleg) / math.sin(total_dogleg)) * t2
        )

        t /= np.linalg.norm(t)

        inc, azi = get_angles(t)[0]

    mult = x / (survey.delta_md[index + 1])

    cov_nev = None if survey.cov_nev is None else (
        survey.cov_nev[index]
        + (
            np.full(shape=(1, 3, 3), fill_value=mult)
            * (survey.cov_nev[index+1] - survey.cov_nev[index])
        )
    ).reshape(3, 3)

    sh = survey.header
    sh.azi_reference = 'grid'

    s = Survey(
        md=np.array(
            [survey.md[index], survey.md[index] + x],
            dtype='object'
        ).astype(np.float64),  # this is to prevent VisibleDeprecationWarning
        inc=np.array([survey.inc_rad[index], inc]),
        azi=np.array([survey.azi_grid_rad[index], azi]),
        cov_nev=(
            None if cov_nev is None
            else np.array([survey.cov_nev[index], cov_nev])
        ),
        start_xyz=np.array([survey.x, survey.y, survey.z]).T[index],
        start_nev=np.array([survey.n, survey.e, survey.tvd]).T[index],
        header=sh,
        deg=False,
    )

    interpolated = False if any((
        x == 0,
        x == survey.md[index + 1] - survey.md[index]
     )) else True
    s.interpolated = [False, interpolated]

    return s


def interpolate_tvd(survey, tvd, **kwargs):
    # only seem to work with relative small delta_md - re-write with minimize
    # function?

    def tidy_up_angle(d):
        """
        Helper function to handle large angles.
        """
        if abs(d) > np.pi:
            d %= (2 * np.pi)
        return d

    coeff = 1
    # find closest point assuming tvd is sorted list
    idx = np.searchsorted(survey.tvd, tvd, side="right") - 1
    if idx == len(survey.tvd) - 1:
        idx = len(survey.tvd) - 2
    elif idx == -1:
        idx = len(survey.tvd) - 2
        coeff = -1
    pos1, pos2 = np.array([survey.n, survey.e, survey.tvd]).T[idx: idx + 2]
    vec1, vec2 = survey.vec_nev[idx: idx + 2]
    dogleg = survey.dogleg[idx + 1]
    delta_md = survey.delta_md[idx + 1]

    node_origin = kwargs.get('node_origin')
    if node_origin:
        pos1, vec1 = node_origin.pos_nev, node_origin.vec_nev
        delta_md = survey.md[idx + 1] - node_origin.md
        # TODO: need to recalculate the dogleg
        s_temp = Survey(
            md=[node_origin.md, survey.md[idx + 1]],
            inc=[node_origin.inc_rad, survey.inc_rad[idx + 1]],
            azi=[node_origin.azi_rad, survey.azi_grid_rad[idx + 1]],
            deg=False
        )
        dogleg = s_temp.dogleg[-1]

    if np.isnan(dogleg):
        return _interpolate_survey(survey, x=0, index=idx)

    if dogleg == 0:
        x = (
            (
                tvd - survey.tvd[idx]
            )
            / (survey.tvd[idx + 1] - survey.tvd[idx])
        ) * (survey.md[idx + 1] - survey.md[idx])
    else:
        m = np.array([0., 0., coeff])
        a = np.dot(m, vec1) * np.sin(dogleg)
        b = np.dot(m, vec1) * np.cos(dogleg) - np.dot(m, vec2)
        # p = get_unit_vec(np.array([0., 0., tvd]) - pos1)
        p = np.array([0., 0., tvd]) - pos1
        c = (
            np.dot(m, p)
            * dogleg
            * np.sin(dogleg)
            / delta_md
        ) + b

        d1 = 2 * np.arctan2(
            (
                a + (a ** 2 + b ** 2 - c ** 2) ** 0.5
            ),
            (
                b + c
            )
        )
        d1 = tidy_up_angle(d1)

        d2 = 2 * np.arctan2(
            (
                a - (a ** 2 + b ** 2 - c ** 2) ** 0.5
            ),
            (
                b + c
            )
        )
        d2 = tidy_up_angle(d2)

        assert d1 >= 0 or d2 >= 0
        if d1 < 0:
            d = d2
        elif d2 < 0:
            d = d1
        else:
            d = min(d1, d2)

        x = d / dogleg * delta_md

        if node_origin:
            x -= survey.md[idx] - node_origin.md

        assert x <= delta_md

    interpolated_survey = _interpolate_survey(survey, x=x, index=idx)

    interpolated = True if x > 0 else False
    node = get_node(interpolated_survey, 1, interpolated=interpolated)

    return node


def slice_survey(survey: Survey, start: int, stop: int = None):
    """
    Take a slice from a welleng.survey.Survey object.

    Parameters
    ----------
    survey: welleng.survey.Survey object
    start: int
        The start index of the desired slice.
    stop: int (default: None)
        The stop index of the desired slice, else the remainder of
        the well bore TD is the default.

    Returns
    -------
    s: welleng.survey.Survey object
        A survey object of the desired slice is returned.
    """
    # Removing this start + 2 code - define this explicitly when making call 
    # if stop is None:
    #     stop = start + 2
    md, inc, azi = survey.survey_rad[start:stop].T
    nevs = np.array([survey.n, survey.e, survey.tvd]).T[start:stop]
    n, e, tvd = nevs.T
    # vec = survey.vec[start:stop]

    # Handle `None` values:
    cov_hla = None if survey.cov_hla is None else survey.cov_hla[start:stop]
    cov_nev = None if survey.cov_nev is None else survey.cov_nev[start:stop]

    s = Survey(
        md=md,
        inc=inc,
        azi=azi,
        n=n,
        e=e,
        tvd=tvd,
        header=survey.header,
        radius=survey.radius[start:stop],
        cov_hla=cov_hla,
        cov_nev=cov_nev,
        start_nev=[n[0], e[0], tvd[0]],
        deg=False,
        unit=survey.unit,
    )

    s.error_model = survey.error_model

    return s


def make_cov(a, b, c, diag=False):
    """
    Make a covariance matrix from the 1-sigma errors.

    Parameters
    ----------
        a: (,n) list or array of floats
            Errors in H or N/y axis.
        b: (,n) list or array of floats
            Errors in L or E/x axis.
        c: (,n) list or array of floats
            Errors in A or V/TVD axis.
        diag: boolean (default=False)
            If true, only the lead diagonal is calculated
            with zeros filling the remainder of the matrix.

    Returns
    -------
        cov: (n,3,3) np.array
    """

    if diag:
        z = np.zeros_like(np.array([a]).reshape(-1))
        cov = np.array([
            [a * a, z, z],
            [z, b * b, z],
            [z, z, c * c]
        ]).T
    else:
        cov = np.array([
            [a * a, a * b, a * c],
            [a * b, b * b, b * c],
            [a * c, b * c, c * c]
        ]).T

    return cov


def make_long_cov(arr):
    """
    Make a covariance matrix from the half covariance 1sigma data.
    """
    aa, ab, ac, bb, bc, cc = np.array(arr).T
    cov = np.array([
        [aa, ab, ac],
        [ab, bb, bc],
        [ac, bc, cc]
    ]).T

    return cov


class SplitSurvey:
    def __init__(
        self,
        survey,
    ):
        self.md1, self.inc1, self.azi1 = survey.survey_rad[:-1].T
        self.md2, self.inc2, self.azi2 = survey.survey_rad[1:].T
        self.delta_azi = self.azi2 - self.azi1
        self.delta_inc = self.inc2 - self.inc1

        self.vec1_xyz = survey.vec_xyz[:-1]
        self.vec1_nev = get_nev(self.vec1_xyz)
        self.vec2_xyz = survey.vec_xyz[1:]
        self.vec2_nev = get_nev(self.vec2_xyz)
        self.dogleg = survey.dogleg[1:]


def get_circle_radius(survey, **targets):
    # TODO: add target data to sections
    ss = SplitSurvey(survey)

    y1, x1, z1 = np.cross(ss.vec1_nev, survey.normals).T
    y2, x2, z2 = np.cross(ss.vec2_nev, survey.normals).T

    b1 = np.array([y1, x1, z1]).T
    b2 = np.array([y2, x2, z2]).T
    nev = np.array([survey.n, survey.e, survey.tvd]).T

    cc1 = (
        nev[:-1] - b1
        / np.linalg.norm(b1, axis=1).reshape(-1, 1)
        * survey.curve_radius[:-1].reshape(-1, 1)
    )
    cc2 = (
        nev[1:] - b2
        / np.linalg.norm(b2, axis=1).reshape(-1, 1)
        * survey.curve_radius[1:].reshape(-1, 1)
    )

    starts = np.vstack((cc1, cc2))
    ends = np.vstack((nev[:-1], nev[1:]))

    # n = 1

    return (starts, ends)


def get_sections(survey, rtol=1e-1, atol=1e-1, dls_cont=False, **targets):
    """
    Tries to discretize a survey file into hold or curve sections. These
    sections can then be used to generate a WellPlan object to generate a
    .wbp format file for import into Landmark COMPASS, thus converting a
    survey file to an editable well trajectory.

    Note that this is in development and only tested on output from planning
    software. In its current form it likely won't be too successful on
    "as drilled" surveys (but optimizing the tolerances may help).

    Parameters
    ----------
    survey: welleng.survey.Survey object
    rtol: float (default: 1e-1)
        The relative tolerance when comparing the normals using the
        numpy.isclose() function.
    atol: float (default: 1e-2)
        The absolute tolerance when comparing the normals using the
        numpy.isclose() function.
    dls_cont: bool
        Whether to explicitly check for dls continuity. May results in a
        larger number of control points but a trajectory that is a closer
        fit to the survey.
    **targets: list of Target objects
        Not supported yet...

    Returns:
    --------
    sections: list of welleng.exchange.wbp.TurnPoint objects
    """
    # it turns out that since the well is being split into "holds" and "turns"
    # that the method can always be "920", since even a hold can be expressed
    # as an [md, inc, azi]. This simplifies things greatly!

    METHOD = "920"  # the COMPASS method for minimum curvature

    # TODO: add target data to sections
    # ss = SplitSurvey(survey)

    # check for DLS continuity
    if not dls_cont:
        # dls_cont = [True] * (len(survey.dls) - 2)
        dls_cont = np.full(len(survey.dls) - 2, True)
    else:
        upper = np.around(survey.dls[1:-1], decimals=2)
        lower = np.around(survey.dls[2:], decimals=2)
        # dls_cont = [
        #     True if u == l else False
        #     for u, l in zip(upper, lower)
        # ]
        dls_cont = np.equal(upper, lower)

    continuous = np.all((
        np.all(
            np.isclose(
                survey.normals[:-1],
                survey.normals[1:],
                rtol=rtol, atol=atol,
                equal_nan=True
            ), axis=-1
        ),
        dls_cont
    ), axis=0)

    starts = np.concatenate((
        np.array([0]),
        np.where(continuous == False)[0] + 1,
        np.array([len(survey.md) - 1])
    ))

    actions = ["hold"]
    actions.extend([
        "hold" if d == 0.0 else "curve"
        for d in survey.dogleg[starts[:-1] + 1]
    ])

    sections = []
    tie_on = True
    # for i, (s, e, a) in enumerate(zip(starts, ends, actions)):
    for i, (s, a) in enumerate(zip(starts, actions)):
        md = survey.md[s]
        inc = survey.inc_deg[s]
        azi = survey.azi_grid_deg[s]
        x = survey.e[s]
        y = survey.n[s]
        z = -survey.tvd[s]
        location = [x, y, z]

        # target = ""
        if survey.unit == 'meters':
            denominator = 30
        else:
            denominator = 100

        if a == "hold" or tie_on or i == 0:
            dls = 0.0
            toolface = 0.0
            build_rate = 0.0
            turn_rate = 0.0
            method = METHOD
        else:
            # COMPASS appears to look back, i.e. at a design point in the
            # well plan it looks back to what the dls and toolface was
            # required to get to that point, so need to give it the data from
            # the previous start point.
            lb = starts[i - 1]
            method = METHOD
            dls = survey.dls[s]
            toolface = abs(np.degrees(survey.toolface[starts[i - 1]]))

            azi_p = sections[-1].azi
            if azi - azi_p < -180:
                coeff = 1
            elif azi - azi_p > 180:
                coeff = -1
            else:
                with np.errstate(all='ignore'):
                    coeff = (azi - azi_p) / abs(azi - azi_p)
            if np.isnan(coeff):
                coeff = 1

            toolface *= coeff

            # looks like the toolface is in range -180 to 180 in the .wbp file
            # toolface = toolface - 360 if toolface > 180 else toolface
            delta_md = md - survey.md[lb]

            # TODO: should sum this line by line to avoid issues with long
            # sections
            build_rate = abs(
                (survey.inc_deg[s] - survey.inc_deg[lb])
                / delta_md * denominator
            )

            # TODO: should sum this line by line to avoid issues with long
            # sections need to be careful with azimuth straddling north
            delta_azi_1 = survey.azi_grid_deg[s] - survey.azi_grid_deg[lb]
            if delta_azi_1 < -180:
                delta_azi_1 += 360
            if delta_azi_1 > 180:
                delta_azi_1 -= 360

            delta_azi_2 = 360 - delta_azi_1
            delta_azi = min(delta_azi_1, delta_azi_2)

            delta_azi = delta_azi_1
            turn_rate = delta_azi / delta_md * denominator

        section = TurnPoint(
            md=md,
            inc=inc,
            azi=azi,
            build_rate=build_rate,
            turn_rate=turn_rate,
            dls=dls,
            toolface=toolface,
            method=method,
            target=None,
            tie_on=tie_on,
            location=location
        )

        sections.append(section)

        # Repeat the first section so that creating .wbp works
        if tie_on:
            section.method = '2'
            sections.append(section)
            sections[-1].tie_on = False

        tie_on = False

    return sections


def get_unit(unit):
    if unit in ['m', 'meters']:
        return 'meters'
    elif unit in ['ft', 'feet']:
        return 'feet'
    else:
        return None


def make_survey_header(data):
    """
    Takes a dictionary of survey header data with the same keys as the
    SurveyHeader class properties and returns a SurveyHeader object.
    """
    sh = SurveyHeader()

    for k, v in data.items():
        setattr(sh, k, v)

    return sh


# def save(survey, filename):
#     """
#     Saves the survey header and survey to a text file.
#     """
#     export_csv(survey, filename)


def export_csv(
    survey, filename, tolerance=0.1, dls_cont=False, decimals=3, **kwargs
):
    """
    Function to export a minimalist (only the control points - i.e. the
    begining and end points of hold and/or turn sections) survey to input into
    third party trajectory planning software.

    Parameters
    ----------
    survey: welleng.survey.Survey object
    filename: str
        The path and filename for saving the text file.
    tolerance: float (default: 0.1)
        How close the the final N, E, TVD position of the minimalist survey
        should be to the original survey point (e.g. within 1 meter)
    dls_cont: bool
        Whether to explicitly check for dls continuity. May result in a
        larger number of control points but a trajectory that is a closer
        fit to the survey.
    decimals: int (default: 3)
        Number of decimal places provided in the output file listing
    """

    start_tol = 0

    res = minimize(
        func, start_tol, args=(survey, dls_cont, tolerance), method='SLSQP',
        bounds=[[0, 1.0]], options={'eps': 0.001}
    )

    data = get_data(
        res.x[0], survey, dls_cont
    )

    headers = ','.join([
        'MD',
        'INC (deg)',
        'AZI (deg)',
        'NORTHING (m)',
        'EASTING (m)',
        'TVDSS (m)',
        'DLS',
        'TOOLFACE',
        'BUILD RATE',
        'TURN RATE'
    ])

    if filename is None:
        try:
            import pandas as pd

            df = pd.DataFrame(
                data,
                columns=headers.split(',')
            )
            return df
        except ImportError:
            print("Missing pandas dependency")

    author = kwargs.get('author', 'Jonny Corcutt')
    comments = [
        f"welleng, version: {__version__}\n"
        f"author, {author}\n"
    ]
    comments.extend([
        f"{k}, {v}\n" for k, v in vars(survey.header).items()
    ])
    comments += "\n"
    comments = ''.join(comments)

    np.savetxt(
        filename,
        data,
        delimiter=',',
        fmt=f"%.{decimals}f",
        header=headers,
        comments=comments
    )


def get_data(tol, survey, dls_cont):

    rtol = atol = tol

    sections = survey._get_sections(rtol=rtol, atol=atol, dls_cont=dls_cont)

    data = [[
        tp.md,
        tp.inc,
        tp.azi,
        tp.location[1],
        tp.location[0],
        tp.location[2],
        tp.dls,
        tp.toolface,
        tp.build_rate,
        tp.turn_rate,
    ] for tp in sections]

    data = np.vstack(data[1:])

    return data


def func(x0, survey, dls_cont, tolerance):

    data = get_data(x0, survey, dls_cont)

    md, inc, azi, n, e, tvd, dls, tf, br, tr = data.T
    nev = np.array([survey.n, survey.e, survey.tvd]).T

    s = Survey(
        md=md,
        inc=inc,
        azi=azi,
        start_nev=nev[0],
        header=survey.header
    )

    s_nev = np.array([s.n, s.e, s.tvd]).T

    diff = abs(
        tolerance - np.amax(np.absolute(s_nev[-1] - nev[-1]))
    )

    return diff


def _remove_duplicates(md, inc, azi):
    arr = np.array([md, inc, azi]).T
    upper = arr[:-1]
    lower = arr[1:]

    temp = np.vstack((
        upper[0],
        lower[lower[:, 0] != upper[:, 0]]
    ))

    return temp.T


def from_connections(
    section_data, step=None, survey_header=None,
    start_nev=[0., 0., 0.],
    start_xyz=[0., 0., 0.],
    start_cov_nev=None,
    radius=10, deg=False, error_model=None,
    depth_unit='meters', surface_unit='meters'
):
    """
    Constructs a well survey from a list of sections of control points.

    Parameters
    ----------
        section_data: list of dicts with section data
        start_nev: (3) array of floats (default: [0,0,0])
            The starting position in NEV coordinates.
        radius: float (default: 10)
            The radius is passed to the `welleng.survey.Survey` object
            and represents the radius of the wellbore. It is also used
            when visualizing the results, so can be used to make the
            wellbore *thicker* in the plot.

    Results
    -------
        survey: `welleng.survey.Survey` object
    """
    if type(section_data) is not list:
        section_data = [section_data]

    # get reference mds
    mds_ref = []
    for s in section_data:
        mds_ref.extend([s.md1, s.md_target])

    section_data_interp = interpolate_well(section_data, step)
    # generate lists for survey
    md, inc, azi = np.vstack([np.array(list(zip(
            s['md'].tolist(),
            s['inc'].tolist(),
            s['azi'].tolist(),
        )))
        for s in section_data_interp
    ]).T

    # remove duplicates
    md, inc, azi = _remove_duplicates(md, inc, azi)

    if survey_header is None:
        survey_header = SurveyHeader(
            depth_unit=depth_unit,
            surface_unit=surface_unit
        )

    interpolated = np.array([False if m in mds_ref else True for m in md])

    survey = Survey(
        md=md,
        inc=inc,
        azi=azi,
        start_nev=section_data[0].pos1 + start_nev,
        start_xyz=start_xyz,
        start_cov_nev=start_cov_nev,
        deg=deg,
        radius=radius,
        header=survey_header,
        error_model=error_model,
        unit=depth_unit,
        interpolated=interpolated
    )

    return survey


def interpolate_survey(survey, step=30, dls=1e-8):
    '''
    Interpolate a sparse survey with the desired md step.

    Parameters
    ----------
    survey: welleng.survey.Survey object
    step: float (default=30)
        The desired delta md between stations.
    dls: float (default=0.01)
        The design DLS used to calculate the minimum curvature. This will be
        the minimum DLS used to fit a curve between stations so should be set
        to a small value to ensure a continuous curve is fit without any
        tangent sections.

    Returns
    -------
    survey_interpolated: welleng.survey.Survey object
        Note that a `interpolated` property is added indicating if the survey
        stations is interpolated (True) or not (False).
    '''
    if survey.header.azi_reference == 'true':
        azi = survey.azi_true_rad
    elif survey.header.azi_reference == 'grid':
        azi = survey.azi_grid_rad
    else:
        azi = survey.azi_mag_rad

    s = np.array([survey.md, survey.inc_rad, azi]).T

    s_upper = s[:-1]
    s_lower = s[1:]
    well = []

    for i, (u, l) in enumerate(zip(s_upper, s_lower)):
        if i == 0:
            node1 = Node(
                pos=survey.start_nev,
                md=u[0],
                inc=u[1],
                azi=u[2],
                degrees=False,
                unit=survey.unit
            )
        else:
            node1 = well[-1].node_end
        node2 = Node(
            md=l[0],
            inc=l[1],
            azi=l[2],
            degrees=False,
            unit=survey.unit
        )
        c = Connector(
            node1=node1,
            node2=node2,
            dls_design=dls,
            degrees=False,
            force_min_curve=True,
            unit=survey.unit
        )
        well.append(c)

    survey_interpolated = from_connections(
        well,
        step=step,
        start_xyz=survey.start_xyz,
        survey_header=survey.header,
        error_model=None
    )

    survey_interpolated.interpolated = [
        False if md in survey.md else True
        for md in survey_interpolated.md
    ]

    i = -1
    radii = []
    cov_nev = []
    for (md, boolean) in zip(
        survey_interpolated.md,
        survey_interpolated.interpolated
    ):
        if not boolean:
            i += 1
            if survey.error_model is not None:
                # interpolate covariance error between survey stations
                j = 1 if i < len(survey.md) - 1 else 0
                delta_md = survey.md[i + j] - survey.md[i]
                delta_cov_nev = (
                    survey.cov_nev[i + j] - survey.cov_nev[i]
                )
                unit_cov_nev = (
                    delta_cov_nev / delta_md
                    if j == 1
                    else 0
                )
        radii.append(survey.radius[i])
        if survey.error_model is not None:
            cov_nev.append(
                survey.cov_nev[i]
                + (
                    (md - survey.md[i]) * unit_cov_nev
                )
            )
    survey_interpolated.radius = np.array(radii)
    if bool(cov_nev):
        survey_interpolated.cov_nev = np.array(cov_nev)
        survey_interpolated.cov_hla = NEV_to_HLA(
            survey_interpolated.survey_rad,
            survey_interpolated.cov_nev.T
        ).T

    return survey_interpolated


def get_node_tvd(survey, node1, node2, tvd, node_origin):
    node2.pos_nev, node2.pos_xyz = None, None
    c = Connector(node1=node1, node2=node2, dls_design=1e-8)
    s = from_connections(c, step=None)
    node_new = interpolate_tvd(s, tvd, node_origin=node_origin)

    return node_new


def interpolate_survey_tvd(survey, start=None, stop=None, step=10):
    """
    """
    tvds = [start] if start is not None else [survey.tvd[0]]
    nodes = []

    for i, md2 in enumerate(survey.md):
        if i == 0:
            nodes.append(get_node(survey, i))
            continue
        node_origin = nodes[-1]
        node2_master = survey.interpolate_md(md2)

        while 1:
            node1 = nodes[-1]
            node2 = copy(node2_master)
            if np.isclose(node1.md, md2):
                node1.interpolated = False
                break

            # check if heading upwards
            if node1.pos_nev[2] > tvds[-1] >= node2.pos_nev[2]:
                tvd = tvds[-1]
                node_new = get_node_tvd(survey, node1, node2, tvd, node_origin)
                node_new.interpolated = True
                nodes.append(node_new)
                tvds.append(tvd - step)
            elif node1.pos_nev[2] < (tvds[-1] + step) <= node2.pos_nev[2]:
                tvd = tvds[-1] + step
                node_new = get_node_tvd(survey, node1, node2, tvd, node_origin)
                node_new.interpolated = True
                nodes.append(node_new)
                tvds.append(tvd)
            else:
                nodes.append(node2_master)
                tvds.append(tvds[-1])
                break

    md, inc, azi, interpolated = np.array([
        [n.md, n.inc_rad, n.azi_rad, n.interpolated]
        for n in nodes
    ]).T

    s_interp = Survey(
        md=md,
        inc=inc,
        azi=azi,
        interpolated=interpolated,
        deg=False,
        header=survey.header
    )

    return s_interp


def project_ahead(pos, vec, delta_md, dls, toolface, md=0.0):
    """
    Apply a simple arc or hold from a current position and vector.

    Parameters
    ----------
    pos: (3) array of floats
        Current position in n, e, tvd coordinates.
    vec: (3) array of floats
        Current vector in n, e, tvd coordinates.
    delta_md: float
        The desired along hole projection length.
    dls: float
        The desired dogleg severity of the projection. Entering 0.0 will
        result in a hold section.
    toolface: float
        The desired toolface for the projection.
    md: float (optional)
        The current md if applicable.

    Returns
    -------
    node: welleng.node.Node object
    """
    if dls > 0:
        radius = radius_from_dls(dls)
        dogleg = np.radians(delta_md / 30 * dls)

        pos_temp = np.array([
            np.cos(dogleg),
            0.,
            np.sin(dogleg)
        ]) * radius
        pos_temp[0] = radius - pos_temp[0]

        vec_temp = np.array([
            np.sin(dogleg),
            0.,
            np.cos(dogleg)
        ])

        inc, azi = get_angles(vec, nev=True).reshape(2)

        angles = [
            toolface,
            inc,
            azi
        ]

        r = R.from_euler('zyz', angles, degrees=False)

        pos_new, vec_new = r.apply(np.vstack((pos_temp, vec_temp)))
        pos_new += pos

    else:
        # if dls is 0 then it's a hold
        pos_new = pos + vec * delta_md
        vec_new = vec

    node = Node(
        pos=pos_new,
        vec=vec_new,
        md=md + delta_md,
    )

    return node


def project_to_target(
    survey,
    node_target,
    dls_design=3.0,
    delta_md=None,
    dls=None, toolface=None,
    step=30,
):
    """
    Project a wellpath from the end of a current survey to a target, taking
    account of the location of the bit relative to the surveying tool if the
    `delta_md` property is not `None`.

    Parameters
    ----------
    survey: welleng.survey.Survey obj
    node_target: welleng.node.Node obj
    dls_design: float
        The dls from which to construct the projected wellpath.
    delta_md: float
        The along hole length from the surveying sensor to the bit.
    dls: float
        The desired dogleg severity for the projection from the survey tool
        to the bit. Entering 0.0 will result in a hold section.
    toolface: float
        The desired toolface for the projection from the survey tool to the
        bit.
    step: float
        The desired survey interval for the projected wellpath to the target.

    Returns
    -------
    node: welleng.survey.Survey obj
    """
    connectors = []
    node_start = Node(
            pos=survey.pos_nev[-1], vec=survey.vec_nev[-1], md=survey.md[-1]
        )
    if dls is None:
        dls = survey.dls[-1]
    if toolface is None:
        toolface = survey.toolface[-1]
    if survey.cov_nev is not None:
        cov_nev = survey.cov_nev[-1]
    else:
        cov_nev = None

    # first project to bit if delta_md is defined
    if delta_md is not None:
        node_bit = project_ahead(
            survey.pos_nev[-1],
            survey.vec_nev[-1],
            delta_md,
            dls,
            toolface,
            survey.md[-1]
        )
        node_bit.pos_nev, node_bit.pos_xyz = None, None
        connectors.append(
            Connector(node_start, node_bit, dls_design=dls_design)
        )
        node_bit = connectors[-1].node_end
    else:
        node_bit = node_start

    connectors.append(
        Connector(
            node_bit, node_target, dls_design
        )
    )
    survey_to_target = from_connections(
        connectors,
        step=step,
        survey_header=survey.header,
        start_cov_nev=cov_nev,
        radius=survey.radius[-1], deg=False, error_model=survey.error_model,
        depth_unit=survey.header.depth_unit,
        surface_unit=survey.header.surface_unit
    )
    return survey_to_target


class SurveyData:
    def __init__(self, survey):
        """
        A class for extracting the minimal amount of data from a `Survey`
        object, with methods for combining data from a list of surveys that
        describe an entire well path.

        Parameters
        ----------
        survey : `welleng.survey.Survey`
        """
        self.header = survey.header
        self.md = survey.md
        self.inc = survey.inc_rad
        self.azi = getattr(
            survey, f"azi_{getattr(self.header, 'azi_reference')}_rad"
        )
        self.start_nev = survey.start_nev
        self.start_xyz = survey.start_xyz
        self.cov_nev = survey.cov_nev
        self.cov_hla = survey.cov_hla
        self.radius = survey.radius

    def append_survey(self, survey):
        """
        Method to extract data from a survey and append it to
        the existing survey data existing in the instance.

        Parameters
        ----------
        survey : `welleng.survey.Survey`
        """
        self.md = np.hstack((self.md, survey.md[1:]))
        self.inc = np.hstack((self.inc, survey.inc_rad[1:]))
        self.azi = np.hstack(
            (
                self.azi,
                getattr(
                    survey, f"azi_{getattr(self.header, 'azi_reference')}_rad"
                )[1:]
            )
        )
        self.cov_nev = np.hstack(
            (
                self.cov_nev.reshape(-1),
                survey.cov_nev[1:].reshape(-1)
            )
        ).reshape(-1, 3, 3)
        self.cov_hla = np.hstack(
            (
                self.cov_hla.reshape(-1),
                survey.cov_hla[1:].reshape(-1)
            )
        ).reshape(-1, 3, 3)
        self.radius = np.hstack((self.radius, survey.radius[1:]))

    def get_survey(self):
        """
        Method to create a `welleng.survey.Survey` object from the survey
        data existing in the instance.

        Returns
        -------
        survey : `welleng.survey.Survey`
        """
        survey = Survey(
            md=self.md,
            inc=self.inc,
            azi=self.azi,
            deg=False,
            start_nev=self.start_nev,
            start_xyz=self.start_xyz,
            cov_nev=self.cov_nev,
            cov_hla=self.cov_hla,
            radius= self.radius,
            header=self.header
        )
        return survey


def splice_surveys(surveys):
    """
    Join together an ordered list of surveys for a well (for example, a list
    of surveys with a different error model for each survey).

    Parameters
    ----------
    surveys : list of `welleng.survey.Survey` objects
        The first survey in the list is assumed to be the shallowest and the
        survey `header` data is taken from this well. Subsequent surveys are
        assumed to be ordered by depth, with the first `md` of the next
        survey being equal to the last `md` of the previous survey.

    Returns
    -------
    spliced_survey : `welleng.survey.Survey` object
        A single survey consisting of the input surveys placed together.

    Notes
    -----
    The returned survey will include the covariance data describing the well
    bore uncertainty, but will not include the error models since these may
    be different for each well section.
    """
    assert type(surveys) is list, "Expected a list of surveys"
    assert type(surveys[0]) is Survey, "Expected a list of surveys"

    for i, s in enumerate(surveys):
        if i == 0:
            survey = SurveyData(s)
            continue
        survey.append_survey(s)

    return survey.get_survey()



try:
    import trimesh
    TRIMESH = True
except ImportError:
    TRIMESH = False
from copy import deepcopy
import numpy as np
from numpy import sin, cos, pi
from scipy.spatial import KDTree

from .utils import HLA_to_NEV, get_sigmas
from .survey import slice_survey, Survey
from .visual import figure


class WellMesh:

    def __init__(
        self,
        survey: Survey,
        n_verts: int = 12,
        sigma: float = 3.0,
        sigma_pa: float = 0.5,
        Sm: float = 0,
        method: str = "ellipse",
    ):
        """
        Create a WellMesh object from a welleng Survey object and a
        welleng Clearance object.

        Parameters
        ----------
            survey: welleng.Survey
            clearance: welleng.Clearance
            n_verts: int
                The number of vertices along the uncertainty ellipse
                edge from which to construct the mesh. Recommended minimum is
                12 and that the number is a multiple of 4.
            sigma: float (default: 3.0)
                The desired standard deviation sigma value of the well bore
                uncertainty.
            sigma_pa: float (default: 0.5)
                The desired "project ahead" value. A remnant of the ISCWSA
                method but may be used in the future to accomodate for well
                bore curvature that is not captured by the mesh.
            Sm: float
                From the ISCWSA method, this is an additional factor applied to
                the well bore radius of the offset well to oversize the hole.
            method: str (default="ellipse")
                The method for constructing the uncertainty edge.
                Either "ellipse", "pedal_curve" or "circle".
        """
        assert TRIMESH, "ImportError: try pip install welleng[easy]"
        self.s = survey
        # self.c = clearance
        self.n_verts = int(n_verts)
        self.sigma = sigma
        self.radius = self.s.radius
        self.Sm = Sm
        self.sigma_pa = sigma_pa

        assert method in ["ellipse", "pedal_curve", "circle"], \
            "Invalid method (ellipse or pedal_curve)"
        self.method = method

        if self.method != 'circle':
            self.sigmaH, self.sigmaL, self.sigmaA = get_sigmas(
                self.s.cov_hla
            )
        self.nevs = np.array([self.s.n, self.s.e, self.s.tvd]).T.reshape(-1, 3)
        self._get_vertices()
        self._align_verts()

        self._get_faces()
        self._make_trimesh()

    # Helper functions #
    def _get_faces(self):
        '''
        Construct a mesh of triangular faces (n,3) on the well bore
        uncertainty edge for the given well bore.
        '''
        total_verts = len(self.vertices.reshape(-1, 3))
        rows = int(total_verts / self.n_verts)

        self.faces = get_faces(self.n_verts, rows)

    def _get_vertices(
        self,
    ):
        '''
        Determine the positions of the vertices on the desired uncertainty
        circumference.
        '''
        if self.method == "circle":
            h = self.s.radius.reshape(-1, 1)
            l = h

        else:
            h = (
                np.array(self.sigmaH) * self.sigma
                + self.radius + self.Sm
                + self.sigma_pa / 2
            ).reshape(-1, 1)

            l = (
                np.array(self.sigmaL) * self.sigma
                + self.radius
                + self.Sm
                + self.sigma_pa / 2
            ).reshape(-1, 1)
            # a = self.s.sigmaA * self.c.k + self.s.radius + self.c.Sm

        if self.method in ["ellipse", "circle"]:
            temp = np.linspace(0, 2 * pi, self.n_verts, endpoint=False)

            temp = np.broadcast_to(temp, (len(self.s.md), len(temp)))
            lam = deepcopy(temp)
            # lam[1:] = (
            #     (temp[1:] + self.s.azi_grid_rad[1:].reshape(-1, 1)) % (2 * np.pi)
            # )
            # theta = np.zeros_like(lam)

            x = h * cos(lam)
            y = l * sin(lam)
            z = np.zeros_like(x)
            vertices = np.stack((x, y, z), axis=-1)

        else:
            # make the vertices evenly spaced around the circumference
            lam = np.concatenate((
                np.arccos(
                    np.linspace(
                        1., -1., int(self.n_verts / 2),
                        endpoint=False
                    )
                ),
                np.arccos(
                    np.linspace(
                        1., -1., int(self.n_verts / 2),
                        endpoint=False)) + np.pi
                )
            )
            f = h * (
                (l ** 2 * np.cos(lam))
                /
                (l ** 2 * (np.cos(lam)) ** 2 + h ** 2 * (np.sin(lam)) ** 2)
            )

            g = l * (
                (h ** 2 * np.sin(lam))
                /
                (l ** 2 * (np.cos(lam)) ** 2 + h ** 2 * (np.sin(lam)) ** 2)
            )
            z = np.zeros_like(f)
            vertices = np.stack((f, g, z), axis=-1)

        vertices = HLA_to_NEV(self.s.survey_rad, vertices, cov=False)

        self.vertices = (
            vertices
            + self.nevs.reshape(tuple(
                (
                    self.nevs.shape[0],
                    1,
                    self.nevs.shape[1]
                )
            ))
        )

    def _get_vertices_vectors(self, vertices, pos_center):
        '''
        Determine the vectors of the vertices relative the the well path
        position.
        '''
        vectors = vertices - pos_center
        normals = np.linalg.norm(vectors, axis=-1)
        with np.errstate(divide='ignore', invalid='ignore'):
            unit_vectors = vectors / normals.reshape(-1, 1)
        unit_vectors[np.where(normals == 0)] = vectors[np.where(normals == 0)]

        return (vectors, normals, unit_vectors)

    def _align_verts(self):
        """
        The meshing gets unstable when the inclination is close to vertical
        (since high side can quickly shift azimuth). This code cycles through
        the ellipses and makes sure that they're all lined up relative to each
        other.
        """
        verts_new_list = []
        for i, verts in enumerate(self.vertices):
            if i == 0:
                verts_new_list.append([verts])
                continue
            else:
                vecs_1, norms_1, units_1 = self._get_vertices_vectors(
                    verts_new_list[-1][0],
                    self.nevs[i - 1]
                )
                vecs_2, norms_2, units_2 = self._get_vertices_vectors(
                    self.vertices[i],
                    self.nevs[i]
                )

                if bool(np.isnan(units_2).any()):
                    raise ValueError('Missing uncertainty data')

                nearest = KDTree(units_2)

                key = units_1[0]
                idx_key = nearest.query(key)[1]

                rotation = units_1[3]
                idx_rotation = nearest.query(rotation)[1]

                # re-order vecs_2 to align with vecs_1
                verts_new = np.vstack((verts[idx_key:], verts[:idx_key]))

                if (
                    (idx_rotation - idx_key) < 0
                    and abs(idx_rotation - idx_key) < len(verts) / 2
                ):
                    verts_new = verts_new[::-1]

                verts_new_list.append([verts_new])

        self.vertices = np.vstack(
            verts_new_list
        ).reshape(self.vertices.shape)

    def _make_trimesh(self):
        '''
        Consturct a trimesh.Trimesh object from the well vertices and faces.
        '''
        vertices = self.vertices.reshape(-1, 3)
        mesh = trimesh.Trimesh(
            vertices=vertices,
            faces=self.faces,
            process=False,
            validate=True,
        )

        # self.mesh = fix_mesh(mesh)
        self.mesh = mesh

    def figure(self, type='mesh3d', **kwargs):
        fig = figure(self, type, **kwargs)
        return fig


def make_trimesh_scene(data):
    """
    Construct a trimesh scene. A collision manager can't be saved, but a scene
    can and a scene can be imported into a collision manager.

    Parameters
    ----------
        data: list of welleng.mesh.WellEng objects.

    Returns
    -------
        scene: a trimesh.scene.scene.Scene object.
    """
    assert TRIMESH, "ImportError: try pip install welleng[easy]"
    scene = trimesh.scene.scene.Scene()
    for well in data:
        mesh = data[well].mesh
        scene.add_geometry(
            mesh, node_name=well, geom_name=well, parent_node_name=None
        )

    return scene


def transform_trimesh_scene(scene, origin=None, scale=100, redux=0.25):
    """
    Transforms a scene by scaling it, reseting the origin/datum and performing
    a reduction in the number of triangles to reduce the file size.

    Parameters
    ----------
        scene: trimesh.scene.scene.Scene object
            A trimesh scene of well meshes
        origin: 3d array [x, y, z]
            The origin of the scene from which the new scene will reset to
            [0, 0, 0].
        scale: float
            A scalar reduction will be performed using this float
        redux: float
            The desired reduction ratio for the number of triangles in each
            mesh.

    Returns
    -------
        scene_transformed: trimesh.scene.scene.Scene object
            A transformed, scaled and reprocessed scene.
    """
    assert TRIMESH, "ImportError: try pip install welleng[easy]"
    i = 0
    if not origin:
        T = np.array([0, 0, 0])
    else:
        T = np.array(origin)
    scene_transformed = trimesh.scene.scene.Scene()
    for well in scene.geometry.items():
        name, mesh = well
        mesh_new = mesh.copy()

        mesh_new.simplify_quadratic_decimation(
            int(len(mesh_new.triangles) * redux)
        )
        mesh_new.vertices -= T

        # change axis convention for visualisation #
        x, y, z = mesh_new.vertices.T
        mesh_new.vertices = (np.array([x, z * -1, y]) / scale).T
        scene_transformed.add_geometry(
            mesh_new, node_name=name, geom_name=name
        )

    return scene_transformed


def sliced_mesh(
    survey,
    n_verts=12,
    sigma=3.0,
    sigma_pa=0.5,
    Sm=0,
    start=0,
    stop=-1,
    step=1,
    method="mesh_ellipse",
):
    """
    Generates a list of mesh objects of a user defined length.

    Parameters
    ----------
        survey: welleng.Survey
        n_verts: int
            The number of vertices along the uncertainty ellipse
            edge from which to construct the mesh. Recommended minimum is
            12 and that the number is a multiple of 4.
        sigma: float (default: 3.0)
            The desired standard deviation sigma value of the well bore
            uncertainty.
        sigma_pa: float (default: 0.5)
            The desired "project ahead" value. A remnant of the ISCWSA method
            but may be used in the future to accommodate for well bore
            curvature that is not captured by the mesh.
        Sm: float
            From the ISCWSA method, this is an additional factor applied to
            the well bore radius of the offset well to oversize the hole.
        method: str (default="ellipse")
            The method for constructing the uncertainty edge.
            Either "ellipse" or "pedal_curve".

    Returns
    -------
        list of trimesh.Trimesh mesh objects.

    """
    meshes = []
    l = len(survey.md)
    i = start
    assert stop <= l, f"Out of range, {stop} > {l}"
    if stop == -1:
        stop = l

    # TODO: set this up for multiprocessing

    while True:
        j = i + step + 1
        if j > stop:
            j = stop
        if j - i < 2 and len(meshes) > 0:
            del meshes[-1]
            i -= step

        # slice a well section and create section survey
        s = slice_survey(survey, i, j)

        # generate a mesh for the section slice
        m = WellMesh(
            s,
            n_verts=n_verts,
            sigma=sigma,
            sigma_pa=sigma_pa,
            Sm=Sm,
            method=method,
        )

        meshes.append(m.mesh)

        if j == stop:
            break
        i += step

    return meshes


def fix_mesh(mesh):
    '''
    For whatever reason, the first pass meshing often results in a mesh
    that is not watertight. This function fixes that, as well as
    repairs the windings and normals.

    Parameters
    ----------
        mesh: trimesh object

    Returns
    -------
        good_mesh: trimesh object
    '''
    assert TRIMESH, "ImportError: try pip install welleng[easy]"
    # it makes two outputs (unique_idx, inverse)
    unique_indices = trimesh.grouping.unique_rows(mesh.faces)[0]
    unique_faces = mesh.faces[unique_indices]

    flat_faces = (
        (unique_faces[:, 0] == unique_faces[:, 1])
        | (unique_faces[:, 0] == unique_faces[:, 2])
        | (unique_faces[:, 1] == unique_faces[:, 2])
    )

    good_faces = unique_faces[~flat_faces]
    good_mesh = trimesh.Trimesh(
        vertices=mesh.vertices, faces=good_faces, process=False
    )

    trimesh.repair.fix_winding(good_mesh)
    trimesh.repair.fix_normals(good_mesh)

    return good_mesh


def get_ends(n_verts, rows):
    """
    Given a number of vertices per station and a number of stations, returns
    the faces of the trimesh vertices.
    """
    B = np.arange(1, n_verts - 1, 1)
    C = np.arange(2, n_verts, 1)
    A = np.zeros_like(B)

    faces = np.array([
        np.array([A, B, C]) + np.array([0]),
        np.array([A, B, C]) + np.array([(rows - 1) * (n_verts)])
    ])

    return faces.transpose((0, 2, 1))


def get_faces(n_verts, rows):
    A = np.arange(n_verts).reshape(-1, 1)
    B = np.roll(A, -1).reshape(-1, 1)

    lower = np.arange(rows) + np.arange(rows) * (n_verts - 1)
    upper = lower[1:]
    lower = lower[:-1]


    faces = np.array([
        A + lower, A + upper, B + upper,
        A + lower, B + lower, B + upper
    ]).T.reshape(-1, 3)

    top, bottom = get_ends(n_verts, rows)

    return np.vstack([top, faces, bottom])



import welleng.clearance
import welleng.io
import welleng.error
import welleng.survey
import welleng.utils
import welleng.mesh
import welleng.visual
import welleng.version
import welleng.errors.tool_errors
import welleng.exchange.wbp
import welleng.exchange.csv
import welleng.target
import welleng.connector
import welleng.exchange.edm
import welleng.fluid
import welleng.node
import welleng.architecture
import welleng.torque_drag
import welleng.units
